"""Type checking and synthesizing code for expressions.

Operates on expressions in a basic block after CFG construction. In particular, we
assume that expressions that involve control flow (i.e. short-circuiting and ternary
expressions) have been removed during CFG construction.

Furthermore, we assume that assignment expressions with the walrus operator := have
been turned into regular assignments and are no longer present. As a result, expressions
are assumed to be side effect free, in the sense that they do not modify the variables
available in the type checking context.

We may alter/desugar AST nodes during type checking. In particular, we turn `ast.Name`
nodes into either `LocalName` or `GlobalName` nodes and `ast.Call` nodes are turned into
`LocalCall` or `GlobalCall` nodes. Furthermore, all nodes in the resulting AST are
annotated with their type.

Expressions can be checked against a given type by the `ExprChecker`, raising a type
error if the expressions doesn't have the expected type. Checking is used for annotated
assignments, return values, and function arguments. Alternatively, the `ExprSynthesizer`
can be used to infer a type for an expression.
"""

import ast
import sys
import traceback
from contextlib import suppress
from dataclasses import replace
from typing import Any, NoReturn, cast

from typing_extensions import assert_never

from guppylang.ast_util import (
    AstNode,
    AstVisitor,
    breaks_in_loop,
    get_type_opt,
    return_nodes_in_ast,
    with_loc,
    with_type,
)
from guppylang.cfg.builder import tmp_vars
from guppylang.checker.core import (
    Context,
    DummyEvalDict,
    FieldAccess,
    Globals,
    Locals,
    Place,
    SetitemCall,
    SubscriptAccess,
    Variable,
)
from guppylang.checker.errors.generic import ExpectedError, UnsupportedError
from guppylang.checker.errors.linearity import NonDroppableForBreakError
from guppylang.checker.errors.py_errors import (
    IllegalPyExpressionError,
    PyExprEvalError,
    PyExprIncoherentListError,
    PyExprNotCPythonError,
    PyExprNotStaticError,
)
from guppylang.checker.errors.type_errors import (
    AttributeNotFoundError,
    BadProtocolError,
    BinaryOperatorNotDefinedError,
    IllegalConstant,
    ModuleMemberNotFoundError,
    NonLinearInstantiateError,
    NotCallableError,
    TypeApplyNotGenericError,
    TypeInferenceError,
    TypeMismatchError,
    UnaryOperatorNotDefinedError,
    WrongNumberOfArgsError,
)
from guppylang.definition.common import Definition
from guppylang.definition.module import ModuleDef
from guppylang.definition.ty import TypeDef
from guppylang.definition.value import CallableDef, ValueDef
from guppylang.error import (
    GuppyError,
    GuppyTypeError,
    GuppyTypeInferenceError,
    InternalGuppyError,
)
from guppylang.experimental import check_function_tensors_enabled, check_lists_enabled
from guppylang.nodes import (
    DesugaredGenerator,
    DesugaredGeneratorExpr,
    DesugaredListComp,
    FieldAccessAndDrop,
    GenericParamValue,
    GlobalName,
    IterEnd,
    IterHasNext,
    IterNext,
    LocalCall,
    MakeIter,
    PartialApply,
    PlaceNode,
    PyExpr,
    SubscriptAccessAndDrop,
    TensorCall,
    TypeApply,
)
from guppylang.span import Span, to_span
from guppylang.tys.arg import TypeArg
from guppylang.tys.builtin import (
    array_type,
    bool_type,
    float_type,
    get_element_type,
    int_type,
    is_array_type,
    is_bool_type,
    is_list_type,
    is_sized_iter_type,
    list_type,
    nat_type,
    option_type,
    string_type,
)
from guppylang.tys.param import ConstParam, TypeParam
from guppylang.tys.parsing import arg_from_ast
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import (
    ExistentialTypeVar,
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    OpaqueType,
    StructType,
    TupleType,
    Type,
    TypeBase,
    function_tensor_signature,
    parse_function_tensor,
    unify,
)

# Mapping from unary AST op to dunder method and display name
unary_table: dict[type[ast.unaryop], tuple[str, str]] = {
    ast.UAdd:   ("__pos__",    "+"),
    ast.USub:   ("__neg__",    "-"),
    ast.Invert: ("__invert__", "~"),
}  # fmt: skip

# Mapping from binary AST op to left dunder method, right dunder method and display name
AstOp = ast.operator | ast.cmpop
binary_table: dict[type[AstOp], tuple[str, str, str]] = {
    ast.Add:      ("__add__",      "__radd__",      "+"),
    ast.Sub:      ("__sub__",      "__rsub__",      "-"),
    ast.Mult:     ("__mul__",      "__rmul__",      "*"),
    ast.Div:      ("__truediv__",  "__rtruediv__",  "/"),
    ast.FloorDiv: ("__floordiv__", "__rfloordiv__", "//"),
    ast.Mod:      ("__mod__",      "__rmod__",      "%"),
    ast.Pow:      ("__pow__",      "__rpow__",      "**"),
    ast.LShift:   ("__lshift__",   "__rlshift__",   "<<"),
    ast.RShift:   ("__rshift__",   "__rrshift__",   ">>"),
    ast.BitOr:    ("__or__",       "__ror__",       "|"),
    ast.BitXor:   ("__xor__",      "__rxor__",      "^"),
    ast.BitAnd:   ("__and__",      "__rand__",      "&"),
    ast.MatMult:  ("__matmul__",   "__rmatmul__",   "@"),
    ast.Eq:       ("__eq__",       "__eq__",        "=="),
    ast.NotEq:    ("__ne__",       "__ne__",        "!="),
    ast.Lt:       ("__lt__",       "__gt__",        "<"),
    ast.LtE:      ("__le__",       "__ge__",        "<="),
    ast.Gt:       ("__gt__",       "__lt__",        ">"),
    ast.GtE:      ("__ge__",       "__le__",        ">="),
}  # fmt: skip


class ExprChecker(AstVisitor[tuple[ast.expr, Subst]]):
    """Checks an expression against a type and produces a new type-annotated AST.

    The type may contain free variables that the checker will try to solve. Note that
    the checker will fail, if some free variables cannot be inferred.
    """

    ctx: Context

    # Name for the kind of term we are currently checking against (used in errors).
    # For example, "argument", "return value", or in general "expression".
    _kind: str

    def __init__(self, ctx: Context) -> None:
        self.ctx = ctx
        self._kind = "expression"

    def _fail(
        self,
        expected: Type,
        actual: ast.expr | Type,
        loc: AstNode | None = None,
    ) -> NoReturn:
        """Raises a type error indicating that the type doesn't match."""
        if not isinstance(actual, TypeBase):
            loc = loc or actual
            _, actual = self._synthesize(actual, allow_free_vars=True)
        if loc is None:
            raise InternalGuppyError("Failure location is required")
        raise GuppyTypeError(TypeMismatchError(loc, expected, actual))

    def check(
        self, expr: ast.expr, ty: Type, kind: str = "expression"
    ) -> tuple[ast.expr, Subst]:
        """Checks an expression against a type.

        The type may have free type variables which will try to be resolved. Returns
        a new desugared expression with type annotations and a substitution with the
        resolved type variables.
        """
        # If we already have a type for the expression, we just have to match it against
        # the target
        if actual := get_type_opt(expr):
            expr, subst, inst = check_type_against(actual, ty, expr, self.ctx, kind)
            if inst:
                expr = with_loc(expr, TypeApply(value=expr, tys=inst))
            return with_type(ty.substitute(subst), expr), subst

        # When checking against a variable, we have to synthesize
        if isinstance(ty, ExistentialTypeVar):
            expr, syn_ty = self._synthesize(expr, allow_free_vars=False)
            return with_type(syn_ty, expr), {ty: syn_ty}

        # Otherwise, invoke the visitor
        old_kind = self._kind
        self._kind = kind or self._kind
        expr, subst = self.visit(expr, ty)
        self._kind = old_kind
        return with_type(ty.substitute(subst), expr), subst

    def _synthesize(
        self, node: ast.expr, allow_free_vars: bool
    ) -> tuple[ast.expr, Type]:
        """Invokes the type synthesiser"""
        return ExprSynthesizer(self.ctx).synthesize(node, allow_free_vars)

    def visit_Constant(self, node: ast.Constant, ty: Type) -> tuple[ast.expr, Subst]:
        act = python_value_to_guppy_type(node.value, node, self.ctx.globals, ty)
        if act is None:
            raise GuppyError(IllegalConstant(node, type(node.value)))
        node, subst, inst = check_type_against(act, ty, node, self.ctx, self._kind)
        assert inst == [], "Const values are not generic"
        return node, subst

    def visit_Tuple(self, node: ast.Tuple, ty: Type) -> tuple[ast.expr, Subst]:
        if not isinstance(ty, TupleType) or len(ty.element_types) != len(node.elts):
            return self._fail(ty, node)
        subst: Subst = {}
        for i, el in enumerate(node.elts):
            node.elts[i], s = self.check(el, ty.element_types[i].substitute(subst))
            subst |= s
        return node, subst

    def visit_List(self, node: ast.List, ty: Type) -> tuple[ast.expr, Subst]:
        check_lists_enabled(node)
        if not is_list_type(ty):
            return self._fail(ty, node)
        el_ty = get_element_type(ty)
        subst: Subst = {}
        for i, el in enumerate(node.elts):
            node.elts[i], s = self.check(el, el_ty.substitute(subst))
            subst |= s
        return node, subst

    def visit_DesugaredListComp(
        self, node: DesugaredListComp, ty: Type
    ) -> tuple[ast.expr, Subst]:
        if not is_list_type(ty):
            return self._fail(ty, node)
        node.generators, node.elt, elt_ty = synthesize_comprehension(
            node, node.generators, node.elt, self.ctx
        )
        subst = unify(get_element_type(ty), elt_ty, {})
        if subst is None:
            actual = list_type(elt_ty)
            return self._fail(ty, actual, node)
        return node, subst

    def visit_Call(self, node: ast.Call, ty: Type) -> tuple[ast.expr, Subst]:
        if len(node.keywords) > 0:
            raise GuppyError(UnsupportedError(node.keywords[0], "Keyword arguments"))
        node.func, func_ty = self._synthesize(node.func, allow_free_vars=False)

        # First handle direct calls of user-defined functions and extension functions
        if isinstance(node.func, GlobalName):
            defn = self.ctx.globals[node.func.def_id]
            if isinstance(defn, CallableDef):
                return defn.check_call(node.args, ty, node, self.ctx)

        # When calling a `PartialApply` node, we just move the args into this call
        if isinstance(node.func, PartialApply):
            node.args = [*node.func.args, *node.args]
            node.func = node.func.func
            return self.visit_Call(node, ty)

        # Otherwise, it must be a function as a higher-order value - something
        # whose type is either a FunctionType or a Tuple of FunctionTypes
        if isinstance(func_ty, FunctionType):
            args, return_ty, inst = check_call(func_ty, node.args, ty, node, self.ctx)
            check_inst(func_ty, inst, node)
            node.func = instantiate_poly(node.func, func_ty, inst)
            return with_loc(node, LocalCall(func=node.func, args=args)), return_ty

        if isinstance(func_ty, TupleType) and (
            function_elements := parse_function_tensor(func_ty)
        ):
            check_function_tensors_enabled(node.func)
            if any(f.parametrized for f in function_elements):
                raise GuppyError(
                    UnsupportedError(node.func, "Polymorphic function tensors")
                )

            tensor_ty = function_tensor_signature(function_elements)

            processed_args, subst, inst = check_call(
                tensor_ty, node.args, ty, node, self.ctx
            )
            assert len(inst) == 0
            return with_loc(
                node,
                TensorCall(func=node.func, args=processed_args, tensor_ty=tensor_ty),
            ), subst

        elif callee := self.ctx.globals.get_instance_func(func_ty, "__call__"):
            return callee.check_call(node.args, ty, node, self.ctx)
        else:
            raise GuppyTypeError(NotCallableError(node.func, func_ty))

    def visit_PyExpr(self, node: PyExpr, ty: Type) -> tuple[ast.expr, Subst]:
        python_val = eval_py_expr(node, self.ctx)
        if act := python_value_to_guppy_type(
            python_val, node.value, self.ctx.globals, ty
        ):
            subst = unify(ty, act, {})
            if subst is None:
                self._fail(ty, act, node)
            act = act.substitute(subst)
            subst = {x: s for x, s in subst.items() if x in ty.unsolved_vars}
            return with_type(act, with_loc(node, ast.Constant(value=python_val))), subst

        raise GuppyError(IllegalPyExpressionError(node.value, type(python_val)))

    def generic_visit(self, node: ast.expr, ty: Type) -> tuple[ast.expr, Subst]:
        # Try to synthesize and then check if we can unify it with the given type
        node, synth = self._synthesize(node, allow_free_vars=False)
        node, subst, inst = check_type_against(synth, ty, node, self.ctx, self._kind)

        # Apply instantiation of quantified type variables
        if inst:
            node = with_loc(node, TypeApply(value=node, inst=inst))

        return node, subst


class ExprSynthesizer(AstVisitor[tuple[ast.expr, Type]]):
    ctx: Context

    def __init__(self, ctx: Context) -> None:
        self.ctx = ctx

    def synthesize(
        self, node: ast.expr, allow_free_vars: bool = False
    ) -> tuple[ast.expr, Type]:
        """Tries to synthesise a type for the given expression.

        Also returns a new desugared expression with type annotations.
        """
        if ty := get_type_opt(node):
            return node, ty
        node, ty = self.visit(node)
        if ty.unsolved_vars and not allow_free_vars:
            raise GuppyError(TypeInferenceError(node, ty))
        return with_type(ty, node), ty

    def _check(
        self, expr: ast.expr, ty: Type, kind: str = "expression"
    ) -> tuple[ast.expr, Subst]:
        """Checks an expression against a given type"""
        return ExprChecker(self.ctx).check(expr, ty, kind)

    def visit_Constant(self, node: ast.Constant) -> tuple[ast.expr, Type]:
        ty = python_value_to_guppy_type(node.value, node, self.ctx.globals)
        if ty is None:
            raise GuppyError(IllegalConstant(node, type(node.value)))
        return node, ty

    def visit_Name(self, node: ast.Name) -> tuple[ast.expr, Type]:
        x = node.id
        if x in self.ctx.locals:
            var = self.ctx.locals[x]
            return with_loc(node, PlaceNode(place=var)), var.ty
        elif x in self.ctx.generic_params:
            param = self.ctx.generic_params[x]
            match param:
                case ConstParam() as param:
                    ast_node = with_loc(node, GenericParamValue(id=x, param=param))
                    return ast_node, param.ty
                case TypeParam() as param:
                    raise GuppyError(
                        ExpectedError(node, "a value", got=f"type `{param.name}`")
                    )
                case _:
                    return assert_never(param)
        elif x in self.ctx.globals:
            defn = self.ctx.globals[x]
            return self._check_global(defn, x, node)
        raise InternalGuppyError(
            f"Variable `{x}` is not defined in `TypeSynthesiser`. This should have "
            "been caught by program analysis!"
        )

    def _check_global(
        self, defn: Definition, name: str, node: ast.expr
    ) -> tuple[ast.expr, Type]:
        """Checks a global definition in an expression position."""
        match defn:
            case ValueDef() as defn:
                return with_loc(node, GlobalName(id=name, def_id=defn.id)), defn.ty
            # For types, we return their `__new__` constructor
            case TypeDef() as defn if constr := self.ctx.globals.get_instance_func(
                defn, "__new__"
            ):
                return with_loc(node, GlobalName(id=name, def_id=constr.id)), constr.ty
            case defn:
                raise GuppyError(
                    ExpectedError(node, "a value", got=f"{defn.description} `{name}`")
                )

    def visit_Attribute(self, node: ast.Attribute) -> tuple[ast.expr, Type]:
        # A `value.attr` attribute access. Unfortunately, the `attr` is just a string,
        # not an AST node, so we have to compute its span by hand. This is fine since
        # linebreaks are not allowed in the identifier following the `.`
        span = to_span(node)
        attr_span = Span(span.end.shift_left(len(node.attr)), span.end)
        if module_def := self._is_module_def(node.value):
            if node.attr not in module_def.globals:
                raise GuppyError(
                    ModuleMemberNotFoundError(attr_span, module_def.name, node.attr)
                )
            defn = module_def.globals[node.attr]
            qual_name = f"{module_def.name}.{defn.name}"
            return self._check_global(defn, qual_name, node)
        node.value, ty = self.synthesize(node.value)
        if isinstance(ty, StructType) and node.attr in ty.field_dict:
            field = ty.field_dict[node.attr]
            expr: ast.expr
            if isinstance(node.value, PlaceNode):
                # Field access on a place is itself a place
                expr = PlaceNode(place=FieldAccess(node.value.place, field, None))
            else:
                # If the struct is not in a place, then there is no way to address the
                # other fields after this one has been projected (e.g. `f().a` makes
                # you loose access to all fields besides `a`).
                expr = FieldAccessAndDrop(value=node.value, struct_ty=ty, field=field)
            return with_loc(node, expr), field.ty
        elif func := self.ctx.globals.get_instance_func(ty, node.attr):
            name = with_type(
                func.ty, with_loc(node, GlobalName(id=func.name, def_id=func.id))
            )
            # Make a closure by partially applying the `self` argument
            # TODO: Try to infer some type args based on `self`
            result_ty = FunctionType(
                func.ty.inputs[1:],
                func.ty.output,
                func.ty.input_names[1:] if func.ty.input_names else None,
                func.ty.params,
            )
            return with_loc(node, PartialApply(func=name, args=[node.value])), result_ty
        raise GuppyTypeError(AttributeNotFoundError(attr_span, ty, node.attr))

    def _is_module_def(self, node: ast.expr) -> ModuleDef | None:
        """Checks whether an AST node corresponds to a defined module."""
        if isinstance(node, ast.Name) and node.id in self.ctx.globals:
            defn = self.ctx.globals[node.id]
            if isinstance(defn, ModuleDef):
                return defn
        return None

    def visit_Tuple(self, node: ast.Tuple) -> tuple[ast.expr, Type]:
        elems = [self.synthesize(elem) for elem in node.elts]

        node.elts = [n for n, _ in elems]
        return node, TupleType([ty for _, ty in elems])

    def visit_List(self, node: ast.List) -> tuple[ast.expr, Type]:
        check_lists_enabled(node)
        if len(node.elts) == 0:
            unsolved_ty = list_type(ExistentialTypeVar.fresh("T", True, True))
            raise GuppyTypeInferenceError(TypeInferenceError(node, unsolved_ty))
        node.elts[0], el_ty = self.synthesize(node.elts[0])
        node.elts[1:] = [self._check(el, el_ty)[0] for el in node.elts[1:]]
        return node, list_type(el_ty)

    def visit_DesugaredListComp(self, node: DesugaredListComp) -> tuple[ast.expr, Type]:
        node.generators, node.elt, elt_ty = synthesize_comprehension(
            node, node.generators, node.elt, self.ctx
        )
        result_ty = list_type(elt_ty)
        return node, result_ty

    def visit_DesugaredGeneratorExpr(
        self, node: DesugaredGeneratorExpr
    ) -> tuple[ast.expr, Type]:
        # This is a generator in an arbitrary expression position. We don't support
        # generators as first-class value yet, so we always error out here. Special
        # cases where generator are allowed need to explicitly check for them (e.g. see
        # the handling of array comprehensions in the compiler for the `array` function)
        raise GuppyError(UnsupportedError(node, "Generator expressions"))

    def visit_UnaryOp(self, node: ast.UnaryOp) -> tuple[ast.expr, Type]:
        # We need to synthesise the argument type, so we can look up dunder methods
        node.operand, op_ty = self.synthesize(node.operand)

        # Special case for the `not` operation since it is not implemented via a dunder
        # method or control-flow
        if isinstance(node.op, ast.Not):
            node.operand, bool_ty = to_bool(node.operand, op_ty, self.ctx)
            return node, bool_ty

        # Check all other unary expressions by calling out to instance dunder methods
        op, display_name = unary_table[node.op.__class__]
        func = self.ctx.globals.get_instance_func(op_ty, op)
        if func is None:
            raise GuppyTypeError(
                UnaryOperatorNotDefinedError(node.operand, op_ty, display_name)
            )
        return func.synthesize_call([node.operand], node, self.ctx)

    def _synthesize_binary(
        self, left_expr: ast.expr, right_expr: ast.expr, op: AstOp, node: ast.expr
    ) -> tuple[ast.expr, Type]:
        """Helper method to compile binary operators by calling out to dunder methods.

        For example, first try calling `__add__` on the left operand. If that fails, try
        `__radd__` on the right operand.
        """
        if op.__class__ not in binary_table:
            raise GuppyTypeError(UnsupportedError(node, "Operator", singular=True))
        lop, rop, display_name = binary_table[op.__class__]
        left_expr, left_ty = self.synthesize(left_expr)
        right_expr, right_ty = self.synthesize(right_expr)

        if func := self.ctx.globals.get_instance_func(left_ty, lop):
            with suppress(GuppyError):
                return func.synthesize_call([left_expr, right_expr], node, self.ctx)

        if func := self.ctx.globals.get_instance_func(right_ty, rop):
            with suppress(GuppyError):
                return func.synthesize_call([right_expr, left_expr], node, self.ctx)

        raise GuppyTypeError(
            # TODO: Is there a way to get the span of the operator?
            BinaryOperatorNotDefinedError(node, left_ty, right_ty, display_name)
        )

    def synthesize_instance_func(
        self,
        node: ast.expr,
        args: list[ast.expr],
        func_name: str,
        description: str,
        exp_sig: FunctionType | None = None,
        give_reason: bool = False,
    ) -> tuple[ast.expr, Type]:
        """Helper method for expressions that are implemented via instance methods.

        Raises a `GuppyTypeError` if the given instance method is not defined. The error
        message can be customised by passing an `err` string and an optional error
        reason can be printed.

        Optionally, the signature of the instance function can also be checked against a
        given expected signature.
        """
        node, ty = self.synthesize(node)
        func = self.ctx.globals.get_instance_func(ty, func_name)
        if func is None:
            err = BadProtocolError(node, ty, description)
            if give_reason and exp_sig is not None:
                err.add_sub_diagnostic(
                    BadProtocolError.MethodMissing(None, func_name, exp_sig)
                )
            raise GuppyTypeError(err)
        if exp_sig and unify(exp_sig, func.ty.unquantified()[0], {}) is None:
            err = BadProtocolError(node, ty, description)
            err.add_sub_diagnostic(
                BadProtocolError.BadSignature(None, ty, func_name, exp_sig, func.ty)
            )
            raise GuppyError(err)
        return func.synthesize_call([node, *args], node, self.ctx)

    def visit_BinOp(self, node: ast.BinOp) -> tuple[ast.expr, Type]:
        return self._synthesize_binary(node.left, node.right, node.op, node)

    def visit_Compare(self, node: ast.Compare) -> tuple[ast.expr, Type]:
        if len(node.comparators) != 1 or len(node.ops) != 1:
            raise InternalGuppyError(
                "BB contains chained comparison. Should have been removed during CFG "
                "construction."
            )
        left_expr, [op], [right_expr] = node.left, node.ops, node.comparators
        return self._synthesize_binary(left_expr, right_expr, op, node)

    def visit_Subscript(self, node: ast.Subscript) -> tuple[ast.expr, Type]:
        node.value, ty = self.synthesize(node.value)
        # Special case for subscripts on functions: Those are type applications
        if isinstance(ty, FunctionType):
            inst = check_type_apply(ty, node, self.ctx)
            return instantiate_poly(node.value, ty, inst), ty.instantiate(inst)
        # Otherwise, it's a regular __getitem__ subscript
        item_expr, item_ty = self.synthesize(node.slice)
        # Give the item a unique name so we can refer to it later in case we also want
        # to compile a call to `__setitem__`
        item = Variable(next(tmp_vars), item_ty, item_expr)
        item_node = with_type(item_ty, with_loc(item_expr, PlaceNode(place=item)))
        # Check a call to the `__getitem__` instance function
        exp_sig = FunctionType(
            [
                FuncInput(ty, InputFlags.Inout),
                FuncInput(
                    ExistentialTypeVar.fresh("Key", True, True), InputFlags.NoFlags
                ),
            ],
            ExistentialTypeVar.fresh("Val", True, True),
        )
        getitem_expr, result_ty = self.synthesize_instance_func(
            node.value, [item_node], "__getitem__", "subscriptable", exp_sig
        )
        # Subscripting a place is itself a place
        expr: ast.expr
        if isinstance(node.value, PlaceNode):
            place = SubscriptAccess(
                node.value.place, item, result_ty, item_expr, getitem_expr
            )
            expr = PlaceNode(place=place)
        else:
            # If the subscript is not on a place, then there is no way to address the
            # other indices after this one has been projected out (e.g. `f()[0]` makes
            # you loose access to all elements besides 0).
            expr = SubscriptAccessAndDrop(
                item=item,
                item_expr=item_expr,
                getitem_expr=getitem_expr,
                original_expr=node,
            )
        return with_loc(node, expr), result_ty

    def visit_Call(self, node: ast.Call) -> tuple[ast.expr, Type]:
        if len(node.keywords) > 0:
            raise GuppyError(UnsupportedError(node.keywords[0], "Keyword arguments"))
        node.func, ty = self.synthesize(node.func)

        # First handle direct calls of user-defined functions and extension functions
        if isinstance(node.func, GlobalName):
            defn = self.ctx.globals[node.func.def_id]
            if isinstance(defn, CallableDef):
                return defn.synthesize_call(node.args, node, self.ctx)

        # When calling a `PartialApply` node, we just move the args into this call
        if isinstance(node.func, PartialApply):
            node.args = [*node.func.args, *node.args]
            node.func = node.func.func
            return self.visit_Call(node)

        # Otherwise, it must be a function as a higher-order value, or a tensor
        if isinstance(ty, FunctionType):
            args, return_ty, inst = synthesize_call(ty, node.args, node, self.ctx)
            node.func = instantiate_poly(node.func, ty, inst)
            return with_loc(node, LocalCall(func=node.func, args=args)), return_ty
        elif isinstance(ty, TupleType) and (
            function_elems := parse_function_tensor(ty)
        ):
            check_function_tensors_enabled(node.func)
            if any(f.parametrized for f in function_elems):
                raise GuppyError(
                    UnsupportedError(node.func, "Polymorphic function tensors")
                )

            tensor_ty = function_tensor_signature(function_elems)
            args, return_ty, inst = synthesize_call(
                tensor_ty, node.args, node, self.ctx
            )
            assert len(inst) == 0

            return with_loc(
                node, TensorCall(func=node.func, args=args, tensor_ty=tensor_ty)
            ), return_ty

        elif f := self.ctx.globals.get_instance_func(ty, "__call__"):
            return f.synthesize_call(node.args, node, self.ctx)
        else:
            raise GuppyTypeError(NotCallableError(node.func, ty))

    def visit_MakeIter(self, node: MakeIter) -> tuple[ast.expr, Type]:
        node.value, ty = self.synthesize(node.value)
        flags = InputFlags.Owned if not ty.copyable else InputFlags.NoFlags
        exp_sig = FunctionType(
            [FuncInput(ty, flags)], ExistentialTypeVar.fresh("Iter", True, True)
        )
        expr, ty = self.synthesize_instance_func(
            node.value, [], "__iter__", "iterable", exp_sig, True
        )
        # Unwrap the size hint if present
        if is_sized_iter_type(ty) and node.unwrap_size_hint:
            expr, ty = self.synthesize_instance_func(expr, [], "unwrap_iter", "")

        # If the iterator was created by a `for` loop, we can add some extra checks to
        # produce nicer errors for linearity violations. Namely, `break` and `return`
        # are not allowed when looping over a non-copyable iterator (`continue` is
        # allowed)
        if not ty.droppable and isinstance(node.origin_node, ast.For):
            breaks = breaks_in_loop(node.origin_node) or return_nodes_in_ast(
                node.origin_node
            )
            if breaks:
                err = NonDroppableForBreakError(breaks[0])
                err.add_sub_diagnostic(
                    NonDroppableForBreakError.NonDroppableIteratorType(node, ty)
                )
                raise GuppyTypeError(err)
        return expr, ty

    def visit_IterHasNext(self, node: IterHasNext) -> tuple[ast.expr, Type]:
        node.value, ty = self.synthesize(node.value)
        flags = InputFlags.Owned if not ty.copyable else InputFlags.NoFlags
        exp_sig = FunctionType([FuncInput(ty, flags)], TupleType([bool_type(), ty]))
        return self.synthesize_instance_func(
            node.value, [], "__hasnext__", "an iterator", exp_sig, True
        )

    def visit_IterNext(self, node: IterNext) -> tuple[ast.expr, Type]:
        node.value, ty = self.synthesize(node.value)
        flags = InputFlags.Owned if not ty.copyable else InputFlags.NoFlags
        exp_sig = FunctionType(
            [FuncInput(ty, flags)],
            option_type(TupleType([ExistentialTypeVar.fresh("T", True, True), ty])),
        )
        return self.synthesize_instance_func(
            node.value, [], "__next__", "an iterator", exp_sig, True
        )

    def visit_IterEnd(self, node: IterEnd) -> tuple[ast.expr, Type]:
        node.value, ty = self.synthesize(node.value)
        flags = InputFlags.Owned if not ty.copyable else InputFlags.NoFlags
        exp_sig = FunctionType([FuncInput(ty, flags)], NoneType())
        return self.synthesize_instance_func(
            node.value, [], "__end__", "an iterator", exp_sig, True
        )

    def visit_ListComp(self, node: ast.ListComp) -> tuple[ast.expr, Type]:
        raise InternalGuppyError(
            "BB contains `ListComp`. Should have been removed during CFG"
            f"construction: `{ast.unparse(node)}`"
        )

    def visit_PyExpr(self, node: PyExpr) -> tuple[ast.expr, Type]:
        python_val = eval_py_expr(node, self.ctx)
        if ty := python_value_to_guppy_type(python_val, node, self.ctx.globals):
            return with_loc(node, ast.Constant(value=python_val)), ty

        raise GuppyError(IllegalPyExpressionError(node.value, type(python_val)))

    def visit_NamedExpr(self, node: ast.NamedExpr) -> tuple[ast.expr, Type]:
        raise InternalGuppyError(
            "BB contains `NamedExpr`. Should have been removed during CFG"
            f"construction: `{ast.unparse(node)}`"
        )

    def visit_BoolOp(self, node: ast.BoolOp) -> tuple[ast.expr, Type]:
        raise InternalGuppyError(
            "BB contains `BoolOp`. Should have been removed during CFG construction: "
            f"`{ast.unparse(node)}`"
        )

    def visit_IfExp(self, node: ast.IfExp) -> tuple[ast.expr, Type]:
        raise InternalGuppyError(
            "BB contains `IfExp`. Should have been removed during CFG construction: "
            f"`{ast.unparse(node)}`"
        )

    def generic_visit(self, node: ast.expr) -> NoReturn:
        """Called if no explicit visitor function exists for a node."""
        raise GuppyError(UnsupportedError(node, "This expression", singular=True))


def check_type_against(
    act: Type, exp: Type, node: ast.expr, ctx: Context, kind: str = "expression"
) -> tuple[ast.expr, Subst, Inst]:
    """Checks a type against another type.

    Returns a substitution for the free variables the expected type and an instantiation
    for the parameters in the actual type. Note that the expected type may not be
    parametrised and the actual type may not contain free unification variables.
    """
    assert not isinstance(exp, FunctionType) or not exp.parametrized
    assert not act.unsolved_vars

    # The actual type may be parametrised. In that case, we have to find an
    # instantiation to avoid higher-rank types.
    subst: Subst | None
    if isinstance(act, FunctionType) and act.parametrized:
        unquantified, free_vars = act.unquantified()
        subst = unify(exp, unquantified, {})
        if subst is None:
            raise GuppyTypeError(TypeMismatchError(node, exp, act, kind))
        # Check that we have found a valid instantiation for all params
        for i, v in enumerate(free_vars):
            param = act.params[i].name
            if v not in subst:
                err = TypeMismatchError(node, exp, act, kind)
                err.add_sub_diagnostic(TypeMismatchError.CantInferParam(None, param))
                raise GuppyTypeInferenceError(err)
            if subst[v].unsolved_vars:
                err = TypeMismatchError(node, exp, act, kind)
                err.add_sub_diagnostic(
                    TypeMismatchError.CantInstantiateFreeVars(None, param, subst[v])
                )
                raise GuppyTypeError(err)
        inst = [subst[v].to_arg() for v in free_vars]
        subst = {v: t for v, t in subst.items() if v in exp.unsolved_vars}

        # Finally, check that the instantiation respects the linearity requirements
        check_inst(act, inst, node)

        return node, subst, inst

    # Otherwise, we know that `act` has no unsolved type vars, so unification is trivial
    assert not act.unsolved_vars
    subst = unify(exp, act, {})
    if subst is None:
        # Maybe we can implicitly coerce `act` to `exp`
        if coerced := try_coerce_to(act, exp, node, ctx):
            return coerced, {}, []
        raise GuppyTypeError(TypeMismatchError(node, exp, act, kind))
    return node, subst, []


def try_coerce_to(
    act: Type, exp: Type, node: ast.expr, ctx: Context
) -> ast.expr | None:
    """Tries to implicitly coerce an expression to a different type.

    Returns the coerced expression or `None` if the type cannot be implicitly coerced.
    """
    # Currently, we only support implicit coercions of numeric types
    if not isinstance(act, NumericType) or not isinstance(exp, NumericType):
        return None
    # Ordering on `NumericType.Kind` defines the coercion relation
    if act.kind < exp.kind:
        f = ctx.globals.get_instance_func(act, f"__{exp.kind.name.lower()}__")
        assert f is not None
        node, subst = f.check_call([node], exp, node, ctx)
        assert len(subst) == 0, "Coercion methods are not generic"
        return node
    return None


def check_type_apply(ty: FunctionType, node: ast.Subscript, ctx: Context) -> Inst:
    """Checks a `f[T1, T2, ...]` type application of a generic function."""
    func = node.value
    arg_exprs = (
        node.slice.elts
        if isinstance(node.slice, ast.Tuple) and len(node.slice.elts) > 0
        else [node.slice]
    )
    globals = ctx.globals

    if not ty.parametrized:
        func_name = globals[func.def_id].name if isinstance(func, GlobalName) else None
        raise GuppyError(TypeApplyNotGenericError(node, func_name))

    exp, act = len(ty.params), len(arg_exprs)
    assert exp > 0
    assert act > 0
    if exp != act:
        if exp < act:
            span = Span(to_span(arg_exprs[exp]).start, to_span(arg_exprs[-1]).end)
        else:
            span = Span(to_span(arg_exprs[-1]).end, to_span(node).end)
        err = WrongNumberOfArgsError(span, exp, act, detailed=True, is_type_apply=True)
        err.add_sub_diagnostic(WrongNumberOfArgsError.SignatureHint(None, ty))
        raise GuppyError(err)

    return [
        param.check_arg(arg_from_ast(arg_expr, globals, ctx.generic_params), arg_expr)
        for arg_expr, param in zip(arg_exprs, ty.params, strict=True)
    ]


def check_num_args(
    exp: int, act: int, node: AstNode, sig: FunctionType | None = None
) -> None:
    """Checks that the correct number of arguments have been passed to a function."""
    if exp == act:
        return
    span, detailed = to_span(node), False
    if isinstance(node, ast.Call):
        # We can construct a nicer error span if we know it's a regular call
        detailed = True
        if exp < act:
            span = Span(to_span(node.args[exp]).start, to_span(node.args[-1]).end)
        elif act > 0:
            span = Span(to_span(node.args[-1]).end, to_span(node).end)
        else:
            span = Span(to_span(node.func).end, to_span(node).end)
    err = WrongNumberOfArgsError(span, exp, act, detailed)
    if sig:
        err.add_sub_diagnostic(WrongNumberOfArgsError.SignatureHint(None, sig))
    raise GuppyTypeError(err)


def type_check_args(
    inputs: list[ast.expr],
    func_ty: FunctionType,
    subst: Subst,
    ctx: Context,
    node: AstNode,
) -> tuple[list[ast.expr], Subst]:
    """Checks the arguments of a function call and infers free type variables.

    We expect that parameters have been replaced with free unification variables.
    Checks that all unification variables can be inferred.
    """
    assert not func_ty.parametrized
    check_num_args(len(func_ty.inputs), len(inputs), node, func_ty)

    new_args: list[ast.expr] = []
    for inp, func_inp in zip(inputs, func_ty.inputs, strict=True):
        a, s = ExprChecker(ctx).check(inp, func_inp.ty.substitute(subst), "argument")
        if InputFlags.Inout in func_inp.flags and isinstance(a, PlaceNode):
            a.place = check_place_assignable(
                a.place, ctx, a, "able to borrow subscripted elements"
            )
        new_args.append(a)
        subst |= s

    # If the argument check succeeded, this means that we must have found instantiations
    # for all unification variables occurring in the input types
    assert all(
        set.issubset(inp.ty.unsolved_vars, subst.keys()) for inp in func_ty.inputs
    )

    # We also have to check that we found instantiations for all vars in the return type
    if not set.issubset(func_ty.output.unsolved_vars, subst.keys()):
        raise GuppyTypeInferenceError(
            TypeInferenceError(node, func_ty.output.substitute(subst))
        )

    return new_args, subst


def check_place_assignable(
    place: Place, ctx: Context, node: ast.expr, reason: str
) -> Place:
    """Performs additional checks for assignments to places, for example for borrowed
    place arguments after function returns.

    In particular, we need to check that places involving `place[item]` subscripts
    implement the corresponding `__setitem__` method.
    """
    match place:
        case Variable():
            return place
        case FieldAccess(parent=parent):
            return replace(
                place, parent=check_place_assignable(parent, ctx, node, reason)
            )
        case SubscriptAccess(parent=parent, item=item, ty=ty):
            # Create temporary variable for the setitem value
            tmp_var = Variable(next(tmp_vars), item.ty, node)
            # Check a call to the `__setitem__` instance function
            exp_sig = FunctionType(
                [
                    FuncInput(parent.ty, InputFlags.Inout),
                    FuncInput(item.ty, InputFlags.NoFlags),
                    FuncInput(ty, InputFlags.Owned),
                ],
                NoneType(),
            )
            setitem_args: list[ast.expr] = [
                with_type(parent.ty, with_loc(node, PlaceNode(parent))),
                with_type(item.ty, with_loc(node, PlaceNode(item))),
                with_type(ty, with_loc(node, PlaceNode(tmp_var))),
            ]
            setitem_call, _ = ExprSynthesizer(ctx).synthesize_instance_func(
                setitem_args[0],
                setitem_args[1:],
                "__setitem__",
                reason,
                exp_sig,
                True,
            )
            return replace(place, setitem_call=SetitemCall(setitem_call, tmp_var))


def synthesize_call(
    func_ty: FunctionType, args: list[ast.expr], node: AstNode, ctx: Context
) -> tuple[list[ast.expr], Type, Inst]:
    """Synthesizes the return type of a function call.

    Returns an annotated argument list, the synthesized return type, and an
    instantiation for the quantifiers in the function type.
    """
    assert not func_ty.unsolved_vars
    check_num_args(len(func_ty.inputs), len(args), node, func_ty)

    # Replace quantified variables with free unification variables and try to infer an
    # instantiation by checking the arguments
    unquantified, free_vars = func_ty.unquantified()
    args, subst = type_check_args(args, unquantified, {}, ctx, node)

    # Success implies that the substitution is closed
    assert all(not t.unsolved_vars for t in subst.values())
    inst = [subst[v].to_arg() for v in free_vars]

    # Finally, check that the instantiation respects the linearity requirements
    check_inst(func_ty, inst, node)

    return args, unquantified.output.substitute(subst), inst


def check_call(
    func_ty: FunctionType,
    inputs: list[ast.expr],
    ty: Type,
    node: AstNode,
    ctx: Context,
    kind: str = "expression",
) -> tuple[list[ast.expr], Subst, Inst]:
    """Checks the return type of a function call against a given type.

    Returns an annotated argument list, a substitution for the free variables in the
    expected type, and an instantiation for the quantifiers in the function type.
    """
    assert not func_ty.unsolved_vars
    check_num_args(len(func_ty.inputs), len(inputs), node, func_ty)

    # When checking, we can use the information from the expected return type to infer
    # some type arguments. However, this pushes errors inwards. For example, given a
    # function `foo: forall T. T -> T`, the following type mismatch would be reported:
    #
    #       x: int = foo(None)
    #                    ^^^^  Expected argument of type `int`, got `None`
    #
    # But the following error location would be more intuitive for users:
    #
    #       x: int = foo(None)
    #                ^^^^^^^^^  Expected expression of type `int`, got `None`
    #
    # In other words, if we can get away with synthesising the call without the extra
    # information from the expected type, we should do that to improve the error.

    # TODO: The approach below can result in exponential runtime in the worst case.
    #  However the bad case, e.g. `x: int = foo(foo(...foo(?)...))`, shouldn't be common
    #  in practice. Can we do better than that?

    # First, try to synthesize
    res: tuple[Type, Inst] | None = None
    try:
        inputs, synth, inst = synthesize_call(func_ty, inputs, node, ctx)
        res = synth, inst
    except GuppyTypeInferenceError:
        pass
    if res is not None:
        synth, inst = res
        subst = unify(ty, synth, {})
        if subst is None:
            raise GuppyTypeError(TypeMismatchError(node, ty, synth, kind))
        return inputs, subst, inst

    # If synthesis fails, we try again, this time also using information from the
    # expected return type
    unquantified, free_vars = func_ty.unquantified()
    subst = unify(ty, unquantified.output, {})
    if subst is None:
        raise GuppyTypeError(TypeMismatchError(node, ty, unquantified.output, kind))

    # Try to infer more by checking against the arguments
    inputs, subst = type_check_args(inputs, unquantified, subst, ctx, node)

    # Also make sure we found an instantiation for all free vars in the type we're
    # checking against
    if not set.issubset(ty.unsolved_vars, subst.keys()):
        unsolved = (subst.keys() - ty.unsolved_vars).pop()
        err = TypeMismatchError(node, ty, func_ty.output.substitute(subst))
        err.add_sub_diagnostic(
            TypeMismatchError.CantInferParam(None, unsolved.display_name)
        )
        raise GuppyTypeInferenceError(err)

    # Success implies that the substitution is closed
    assert all(not t.unsolved_vars for t in subst.values())
    inst = [subst[v].to_arg() for v in free_vars]
    subst = {v: t for v, t in subst.items() if v in ty.unsolved_vars}

    # Finally, check that the instantiation respects the linearity requirements
    check_inst(func_ty, inst, node)

    return inputs, subst, inst


def check_inst(func_ty: FunctionType, inst: Inst, node: AstNode) -> None:
    """Checks if an instantiation is valid.

    Makes sure that the linearity requirements are satisfied.
    """
    for param, arg in zip(func_ty.params, inst, strict=True):
        # Give a more informative error message for linearity issues
        if (
            isinstance(param, TypeParam)
            and isinstance(arg, TypeArg)
            and arg.ty.linear
            and not param.can_be_linear
        ):
            raise GuppyTypeError(
                NonLinearInstantiateError(node, param, func_ty, arg.ty)
            )
        # For everything else, we fall back to the default checking implementation
        param.check_arg(arg, node)


def instantiate_poly(node: ast.expr, ty: FunctionType, inst: Inst) -> ast.expr:
    """Instantiates quantified type arguments in a function."""
    assert len(ty.params) == len(inst)
    if len(inst) > 0:
        node = with_loc(node, TypeApply(value=with_type(ty, node), inst=inst))
        return with_type(ty.instantiate(inst), node)
    return with_type(ty, node)


def to_bool(node: ast.expr, node_ty: Type, ctx: Context) -> tuple[ast.expr, Type]:
    """Tries to turn a node into a bool"""
    if is_bool_type(node_ty):
        return node, node_ty
    synth = ExprSynthesizer(ctx)
    exp_sig = FunctionType([FuncInput(node_ty, InputFlags.Inout)], bool_type())
    return synth.synthesize_instance_func(node, [], "__bool__", "truthy", exp_sig, True)


def synthesize_comprehension(
    node: AstNode, gens: list[DesugaredGenerator], elt: ast.expr, ctx: Context
) -> tuple[list[DesugaredGenerator], ast.expr, Type]:
    """Helper function to synthesise the element type of a list comprehension."""
    # If there are no more generators left, we can check the list element
    if not gens:
        elt, elt_ty = ExprSynthesizer(ctx).synthesize(elt)
        return gens, elt, elt_ty

    # Check the first generator
    gen, *gens = gens
    gen, inner_ctx = check_generator(gen, ctx)

    # Check remaining generators in inner context
    gens, elt, elt_ty = synthesize_comprehension(node, gens, elt, inner_ctx)

    return [gen, *gens], elt, elt_ty


def check_generator(
    gen: DesugaredGenerator, ctx: Context
) -> tuple[DesugaredGenerator, Context]:
    """Helper function to check a single generator.

    Returns the type annotated generator together with a new nested context in which the
    generator variables are bound.
    """
    from guppylang.checker.stmt_checker import StmtChecker

    # Check the iterator in the outer context
    gen.iter_assign = StmtChecker(ctx).visit_Assign(gen.iter_assign)

    # The rest is checked in a new nested context to ensure that variables don't escape
    # their scope
    inner_locals: Locals[str, Variable] = Locals({}, parent_scope=ctx.locals)
    inner_ctx = Context(ctx.globals, inner_locals, ctx.generic_params)
    expr_sth, stmt_chk = ExprSynthesizer(inner_ctx), StmtChecker(inner_ctx)
    gen.iter, iter_ty = expr_sth.visit(gen.iter)
    gen.iter = with_type(iter_ty, gen.iter)

    # The type returned by `next_call` is `Option[tuple[elt_ty, iter_ty]]`
    gen.next_call, option_ty = expr_sth.synthesize(gen.next_call)
    next_ty = get_element_type(option_ty)
    assert isinstance(next_ty, TupleType)
    [elt_ty, _] = next_ty.element_types
    gen.target = stmt_chk._check_assign(gen.target, gen.next_call, elt_ty)

    # Check `if` guards
    for i in range(len(gen.ifs)):
        gen.ifs[i], if_ty = expr_sth.synthesize(gen.ifs[i])
        gen.ifs[i], _ = to_bool(gen.ifs[i], if_ty, inner_ctx)

    return gen, inner_ctx


def eval_py_expr(node: PyExpr, ctx: Context) -> Any:
    """Evaluates a `py(...)` expression."""
    # The method we used for obtaining the Python variables in scope only works in
    # CPython (see `get_py_scope()`).
    if sys.implementation.name != "cpython":
        raise GuppyError(PyExprNotCPythonError(node))

    try:
        python_val = eval(  # noqa: S307
            ast.unparse(node.value),
            None,
            DummyEvalDict(ctx, node.value),
        )
    except DummyEvalDict.GuppyVarUsedError as e:
        raise GuppyError(PyExprNotStaticError(e.node or node, e.var)) from None
    except Exception as e:
        # Remove the top frame pointing to the `eval` call from the stack trace
        tb = e.__traceback__.tb_next if e.__traceback__ else None
        tb_formatted = "".join(traceback.format_exception(type(e), e, tb))
        raise GuppyError(PyExprEvalError(node.value, tb_formatted)) from e
    return python_val


def python_value_to_guppy_type(
    v: Any, node: ast.expr, globals: Globals, type_hint: Type | None = None
) -> Type | None:
    """Turns a primitive Python value into a Guppy type.

    Accepts an optional `type_hint` for the expected expression type that is used to
    infer a more precise type (e.g. distinguishing between `int` and `nat`). Note that
    invalid hints are ignored, i.e. no user error are emitted.

    Returns `None` if the Python value cannot be represented in Guppy.
    """
    match v:
        case bool():
            return bool_type()
        case str():
            return string_type()
        # Only resolve `int` to `nat` if the user specifically asked for it
        case int(n) if type_hint == nat_type() and n >= 0:
            return nat_type()
        # Otherwise, default to `int` for consistency with Python
        case int():
            return int_type()
        case float():
            return float_type()
        case tuple(elts):
            hints = (
                type_hint.element_types
                if isinstance(type_hint, TupleType)
                else len(elts) * [None]
            )
            tys = [
                python_value_to_guppy_type(elt, node, globals, hint)
                for elt, hint in zip(elts, hints, strict=False)
            ]
            if any(ty is None for ty in tys):
                return None
            return TupleType(cast(list[Type], tys))
        case list():
            return _python_list_to_guppy_type(v, node, globals, type_hint)
        case _:
            return None


def _python_list_to_guppy_type(
    vs: list[Any], node: ast.expr, globals: Globals, type_hint: Type | None
) -> OpaqueType | None:
    """Turns a Python list into a Guppy type.

    Returns `None` if the list contains different types or types that are not
    representable in Guppy.
    """
    if len(vs) == 0:
        return array_type(ExistentialTypeVar.fresh("T", True, True), 0)

    # All the list elements must have a unifiable types
    v, *rest = vs
    elt_hint = (
        get_element_type(type_hint) if type_hint and is_array_type(type_hint) else None
    )
    el_ty = python_value_to_guppy_type(v, node, globals, elt_hint)
    if el_ty is None:
        return None
    for v in rest:
        ty = python_value_to_guppy_type(v, node, globals, elt_hint)
        if ty is None:
            return None
        if (subst := unify(ty, el_ty, {})) is None:
            raise GuppyError(PyExprIncoherentListError(node))
        el_ty = el_ty.substitute(subst)
    return array_type(el_ty, len(vs))



from abc import ABC

from hugr import Wire, ops
from hugr import tys as ht
from hugr import val as hv

from guppylang.definition.custom import CustomCallCompiler, CustomInoutCallCompiler
from guppylang.definition.value import CallReturnWires
from guppylang.error import InternalGuppyError
from guppylang.std._internal.compiler.prelude import build_expect_none, build_unwrap
from guppylang.tys.arg import TypeArg


class OptionCompiler(CustomInoutCallCompiler, ABC):
    """Abstract base class for compilers for `Option` methods."""

    @property
    def option_ty(self) -> ht.Option:
        match self.type_args:
            case [TypeArg(ty)]:
                return ht.Option(ty.to_hugr())
            case _:
                raise InternalGuppyError("Invalid type args for Option op")


class OptionConstructor(OptionCompiler, CustomCallCompiler):
    """Compiler for the `Option` constructors `nothing` and `some`."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile(self, args: list[Wire]) -> list[Wire]:
        return [self.builder.add_op(ops.Tag(self.tag, self.option_ty), *args)]


class OptionTestCompiler(OptionCompiler):
    """Compiler for the `Option.is_nothing` and `Option.is_some` methods."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [opt] = args
        cond = self.builder.add_conditional(opt)
        for i in [0, 1]:
            with cond.add_case(i) as case:
                val = hv.TRUE if i == self.tag else hv.FALSE
                opt = case.add_op(ops.Tag(i, self.option_ty), *case.inputs())
                case.set_outputs(case.load(val), opt)
        [res, opt] = cond.outputs()
        return CallReturnWires(regular_returns=[res], inout_returns=[opt])


class OptionUnwrapCompiler(OptionCompiler, CustomCallCompiler):
    """Compiler for the `Option.unwrap` method."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        [opt] = args
        err = "Option.unwrap: value is `Nothing`"
        return list(build_unwrap(self.builder, opt, err).outputs())


class OptionUnwrapNothingCompiler(OptionCompiler, CustomCallCompiler):
    """Compiler for the `Option.unwrap_nothing` method."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        [opt] = args
        err = "Option.unwrap: value is `Some`"
        return list(build_expect_none(self.builder, opt, err).outputs())



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.tys.ty import NoneType


module = GuppyModule("test")


@guppy.type(NoneType(), module=module)
class MyType:
    """A non-iterable type."""


@guppy(module)
def test(x: MyType) -> None:
    for _ in x:
        pass


module.compile()



import pytest

from guppylang.std.builtins import result, nat, array
from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def main(x: int) -> None:
        result("foo", x)

    validate(main)


def test_multi(validate):
    @compile_guppy
    def main(w: nat, x: int, y: float, z: bool) -> None:
        result("a", w)
        result("b", x)
        result("c", y)
        result("d", z)

    validate(main)


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/631")
def test_array(validate):
    @compile_guppy
    def main(
        w: array[nat, 42], x: array[int, 5], y: array[float, 1], z: array[bool, 0]
    ) -> None:
        result("a", w)
        result("b", x)
        result("c", y)
        result("d", z)

    validate(main)


def test_same_tag(validate):
    @compile_guppy
    def main(x: int, y: float, z: bool) -> None:
        result("foo", x)
        result("foo", y)
        result("foo", z)

    validate(main)

def test_py_tag(validate):
    @compile_guppy
    def main(x: int) -> None:
        result(py("a" + "b"), x)

    validate(main)



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array


module = GuppyModule("test")
module.load(quantum)


@guppy(module)
def main() -> array[int, 1]:
    array(False)


module.compile()


def test_import_guppy(benchmark):
    def setup_guppy():
        import guppylang.std.quantum_functional as qf  # noqa: F401
        from guppylang.decorator import guppy
        from guppylang.module import GuppyModule  # noqa: F401
        from guppylang.std import quantum  # noqa: F401
        from guppylang.std.angles import angle, pi  # noqa: F401
        from guppylang.std.builtins import array, py  # noqa: F401
        from guppylang.std.quantum import cx, discard_array, h, qubit  # noqa: F401

        @guppy
        def main() -> None:
            return

    benchmark(setup_guppy)



import functools
import os
import sys
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, TypeVar, cast

if TYPE_CHECKING:
    from guppylang.diagnostic import Error, Fatal


@dataclass
class GuppyError(Exception):
    """An error that occurs during compilation."""

    error: "Error | Fatal"


class GuppyTypeError(GuppyError):
    """Special Guppy exception for type errors."""


class GuppyTypeInferenceError(GuppyError):
    """Special Guppy exception for type inference errors."""


class MissingModuleError(Exception):
    """Special Guppy exception for operations that require a guppy module."""


class InternalGuppyError(Exception):
    """Exception for internal problems during compilation."""


ExceptHook = Callable[[type[BaseException], BaseException, TracebackType | None], Any]


@contextmanager
def exception_hook(hook: ExceptHook) -> Iterator[None]:
    """Sets a custom `excepthook` for the scope of a 'with' block."""
    try:
        # Check if we're inside a jupyter notebook since it uses its own exception
        # hook. If we're in a regular interpreter, this line will raise a `NameError`
        ipython_shell = get_ipython()  # type: ignore[name-defined]

        def ipython_excepthook(
            shell: Any,
            etype: type[BaseException],
            value: BaseException,
            tb: TracebackType | None,
            tb_offset: Any = None,
        ) -> Any:
            return hook(etype, value, tb)

        ipython_shell.set_custom_exc((GuppyError,), ipython_excepthook)
        yield
        ipython_shell.set_custom_exc((), None)
    except NameError:
        pass
    else:
        return

    # Otherwise, override the regular sys.excepthook
    old_hook = sys.excepthook
    sys.excepthook = hook
    yield
    sys.excepthook = old_hook


FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def pretty_errors(f: FuncT) -> FuncT:
    """Decorator to print custom error banners when a `GuppyError` occurs."""

    def hook(
        excty: type[BaseException], err: BaseException, traceback: TracebackType | None
    ) -> None:
        """Custom `excepthook` that intercepts `GuppyExceptions` for pretty printing."""
        if isinstance(err, GuppyError):
            from guppylang.decorator import guppy
            from guppylang.diagnostic import DiagnosticsRenderer

            renderer = DiagnosticsRenderer(guppy._sources)
            renderer.render_diagnostic(err.error)
            sys.stderr.write("\n".join(renderer.buffer))
            sys.stderr.write("\n\nGuppy compilation failed due to 1 previous error\n")
            return

        # If it's not a GuppyError, fall back to default hook
        sys.__excepthook__(excty, err, traceback)

    @functools.wraps(f)
    def pretty_errors_wrapped(*args: Any, **kwargs: Any) -> Any:
        with exception_hook(hook):
            return f(*args, **kwargs)

    return cast(FuncT, pretty_errors_wrapped)


def _pytest_running() -> bool:
    """Checks if we are currently running pytest.

    See https://docs.pytest.org/en/latest/example/simple.html#pytest-current-test-environment-variable
    """
    return "PYTEST_CURRENT_TEST" in os.environ



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit, measure


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct1:
    x: "MyStruct2"


@guppy.struct(module)
class MyStruct2:
    q: qubit


@guppy.declare(module)
def use(s: MyStruct1 @owned) -> None: ...


@guppy(module)
def foo(s: MyStruct1 @owned) -> MyStruct1:
    use(s)
    return s


module.compile()



from pathlib import Path
from typing import Generic, no_type_check

from guppylang.decorator import ModuleIdentifier, guppy
from guppylang.std import builtins
from guppylang.std._internal.compiler.option import (
    OptionConstructor,
    OptionTestCompiler,
    OptionUnwrapCompiler,
    OptionUnwrapNothingCompiler,
)
from guppylang.std.builtins import L, mem_swap, owned
from guppylang.tys.builtin import option_type_def

builtins_module = guppy.get_module(
    ModuleIdentifier(Path(builtins.__file__), "builtins", builtins)
)


@guppy.extend_type(option_type_def, module=builtins_module)
class Option(Generic[L]):  # type: ignore[misc]
    """Represents an optional value."""

    @guppy.custom(OptionTestCompiler(0), module=builtins_module)
    @no_type_check
    def is_nothing(self: "Option[L]") -> bool:
        """Returns `True` if the option is a `nothing` value."""

    @guppy.custom(OptionTestCompiler(1), module=builtins_module)
    @no_type_check
    def is_some(self: "Option[L]") -> bool:
        """Returns `True` if the option is a `some` value."""

    @guppy.custom(OptionUnwrapCompiler(), module=builtins_module)
    @no_type_check
    def unwrap(self: "Option[L]" @ owned) -> L:
        """Returns the contained `some` value, consuming `self`.

        Panics if the option is a `nothing` value.
        """

    @guppy.custom(OptionUnwrapNothingCompiler(), module=builtins_module)
    @no_type_check
    def unwrap_nothing(self: "Option[L]" @ owned) -> None:
        """Returns `None` if the option is a `nothing` value, consuming `self`.

        Panics if the option is a `some` value.
        """

    @guppy(builtins_module)
    @no_type_check
    def take(self: "Option[L]") -> "Option[L]":
        """Swaps the value of `self` with `nothing` and returns the original value."""
        n: Option[L] = nothing()  # type: ignore[type-arg, valid-type]
        mem_swap(n, self)
        return n


@guppy.custom(OptionConstructor(0), module=builtins_module)
@no_type_check
def nothing() -> Option[L]:
    """Constructs a `nothing` optional value."""


@guppy.custom(OptionConstructor(1), module=builtins_module)
@no_type_check
def some(value: L @ owned) -> Option[L]:
    """Constructs a `some` optional value."""



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

module = GuppyModule("test")


@guppy.struct(module)
class MyStruct:
    x: int
    y: bool


@guppy(module)
def foo() -> MyStruct:
    return MyStruct(42, False)


@guppy(module)
def bar() -> None:
    foo().x += 1


module.compile()



from importlib.util import find_spec

import pathlib
import pytest

from tests.error.util import run_error_test


tket2_installed = find_spec("tket2") is not None


path = pathlib.Path(__file__).parent.resolve() / "py_errors"
files = [
    x
    for x in path.iterdir()
    if x.is_file()
    and x.suffix == ".py"
    and x.name not in ("__init__.py", "tket2_not_installed.py")
]

# Turn paths into strings, otherwise pytest doesn't display the names
files = [str(f) for f in files]


@pytest.mark.parametrize("file", files)
@pytest.mark.skipif(not tket2_installed, reason="tket2 is not installed")
def test_py_errors(file, capsys, snapshot):
    run_error_test(file, capsys, snapshot)


@pytest.mark.skipif(tket2_installed, reason="tket2 is installed")
def test_tket2_not_installed(capsys, snapshot):
    path = (
        pathlib.Path(__file__).parent.resolve() / "py_errors" / "tket2_not_installed.py"
    )
    run_error_test(str(path), capsys, snapshot)



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def new_qubit() -> qubit:
    ...


@guppy.declare(module)
def measure(q: qubit @owned) -> bool:
    ...


@guppy(module)
def foo(i: int) -> bool:
    b = False
    while True:
        q = new_qubit()
        if i == 0:
            break
        i -= 1
        b &= measure(q)
    return b


module.compile()



# Configuration file for the Sphinx documentation builder.
# See https://www.sphinx-doc.org/en/master/usage/configuration.html

import inspect
from types import GenericAlias, UnionType

import guppylang

project = "Guppy Compiler"
copyright = "2024, Quantinuum"
author = "Quantinuum"

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

html_theme = "sphinx_book_theme"

html_title = "Guppy compiler development docs"

html_theme_options = {
    "repository_url": "https://github.com/CQCL/guppylang",
    "use_repository_button": True,
    "navigation_with_keys": True,
    "logo": {
        "image_light": "_static/Quantinuum_logo_black.png",
        "image_dark": "_static/Quantinuum_logo_white.png",
    },
}

html_static_path = ["../_static"]
html_css_files = ["custom.css"]

autosummary_generate = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}


# add_module_names = False


def _is_type_alias(x):
    """Checks if an object is a type alias.

    Note that this function only catches non-trivial aliases like `IntList = list[int]`
    or `Number = int | float`. Aliases like `MyAlias = int` cannot easily be detected.
    """
    return isinstance(x, GenericAlias | UnionType)


def _find_aliases(module):
    """Finds all type aliases defined in a guppylang module and its submodules.

    Returns a mapping from alias names to the  module in which they are defined.
    """
    aliases = {}
    for name, x in inspect.getmembers(module):
        if _is_type_alias(x):
            aliases[name] = module
        if inspect.ismodule(x) and x.__name__.startswith("guppylang."):
            aliases |= _find_aliases(x)
    return aliases


# Generate a mapping from type aliases to their qualified name to ensure that autodoc
# doesn't unfold them
_aliases = _find_aliases(guppylang)
autodoc_type_aliases = {
    alias: f"~{module.__name__}.{alias}" for alias, module in _aliases.items()
}

# Also create a set of all qualified alias names
_qualified_aliases = {
    f"{module.__name__}.{alias}" for alias, module in _aliases.items()
}


def resolve_type_aliases(app, env, node, contnode):
    """Resolve :class: references to our type aliases as :data: instead.

    When trying to resolve references in type annotations, Sphinx only looks at :class:
    nodes since types are typically classes. However, type aliases are :data: so are
    not found by default. See https://github.com/sphinx-doc/sphinx/issues/10785
    """
    if (
        node["refdomain"] == "py"
        and node["reftype"] == "class"
        and node["reftarget"] in _qualified_aliases
    ):
        ref = app.env.get_domain("py").resolve_xref(
            env, node["refdoc"], app.builder, "data", node["reftarget"], node, contnode
        )
        return ref


def setup(app):
    app.connect("missing-reference", resolve_type_aliases)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from tests.util import compile_guppy


def test_factorial(validate):
    @compile_guppy
    def factorial1(x: int) -> int:
        acc = 1
        while x > 0:
            acc *= x
            x -= 1
        return acc

    @compile_guppy
    def factorial2(x: int) -> int:
        if x == 0:
            return 1
        return factorial2(x - 1) * x

    @compile_guppy
    def factorial3(x: int, acc: int) -> int:
        if x == 0:
            return acc
        return factorial3(x - 1, acc * x)

    validate(factorial1, name="factorial1")
    validate(factorial2, name="factorial2")
    validate(factorial3, name="factorial3")


def test_even_odd(validate):
    module = GuppyModule("module")

    @guppy(module)
    def is_even(x: int) -> bool:
        if x == 0:
            return True
        return is_odd(x - 1)

    @guppy(module)
    def is_odd(x: int) -> bool:
        if x == 0:
            return False
        return is_even(x - 1)

    validate(module.compile())



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array


module = GuppyModule("test")
module.load(quantum)


@guppy(module)
def main() -> array[int, 50]:
    return array(0 for _ in range(10) for _ in range(5))


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

module = GuppyModule("test")

@guppy.declare(module)
def foo(x: int) -> int:
    ...

@guppy(module)
def main() -> int:
    return (foo, foo)(42)

module.compile()



import ast
from dataclasses import dataclass, field
from typing import ClassVar

from hugr import Node, Wire
from hugr.build import function as hf
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang.ast_util import AstNode, has_empty_body, with_loc, with_type
from guppylang.checker.core import Context, Globals, PyScope
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import check_signature
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.definition.common import CompilableDef, ParsableDef
from guppylang.definition.function import (
    PyFunc,
    compile_call,
    load_with_args,
    parse_py_func,
)
from guppylang.definition.value import CallableDef, CallReturnWires, CompiledCallableDef
from guppylang.diagnostic import Error
from guppylang.error import GuppyError
from guppylang.nodes import GlobalCall
from guppylang.span import SourceMap
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import Type


@dataclass(frozen=True)
class BodyNotEmptyError(Error):
    title: ClassVar[str] = "Unexpected function body"
    span_label: ClassVar[str] = "Body of declared function `{name}` must be empty"
    name: str


@dataclass(frozen=True)
class RawFunctionDecl(ParsableDef):
    """A raw function declaration provided by the user.

    The raw declaration stores exactly what the user has written (i.e. the AST), without
    any additional checking or parsing.
    """

    python_func: PyFunc
    python_scope: PyScope = field(repr=False)
    description: str = field(default="function", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "CheckedFunctionDecl":
        """Parses and checks the user-provided signature of the function."""
        func_ast, docstring = parse_py_func(self.python_func, sources)
        ty = check_signature(func_ast, globals.with_python_scope(self.python_scope))
        if not has_empty_body(func_ast):
            raise GuppyError(BodyNotEmptyError(func_ast.body[0], self.name))
        return CheckedFunctionDecl(
            self.id,
            self.name,
            func_ast,
            ty,
            self.python_func,
            self.python_scope,
            docstring,
        )


@dataclass(frozen=True)
class CheckedFunctionDecl(RawFunctionDecl, CompilableDef, CallableDef):
    """A function declaration with parsed and checked signature.

    In particular, this means that we have determined a type for the function.
    """

    defined_at: ast.FunctionDef
    docstring: str | None

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[GlobalCall, Type]:
        """Synthesizes the return type of a function call."""
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return with_type(ty, node), ty

    def compile_outer(self, module: DefinitionBuilder[OpVar]) -> "CompiledFunctionDecl":
        """Adds a Hugr `FuncDecl` node for this function to the Hugr."""
        assert isinstance(
            module, hf.Module
        ), "Functions can only be declared in modules"
        module: hf.Module = module

        node = module.declare_function(self.name, self.ty.to_hugr_poly())
        return CompiledFunctionDecl(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.python_func,
            self.python_scope,
            self.docstring,
            node,
        )


@dataclass(frozen=True)
class CompiledFunctionDecl(CheckedFunctionDecl, CompiledCallableDef):
    """A function declaration with a corresponding Hugr node."""

    declaration: Node

    def load_with_args(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the function as a value into a local Hugr dataflow graph."""
        # Use implementation from function definition.
        return load_with_args(type_args, dfg, self.ty, self.declaration)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        # Use implementation from function definition.
        return compile_call(args, type_args, dfg, self.ty, self.declaration)



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def foo(qs: array[qubit, 42] @owned, q: qubit) -> array[qubit, 42]: ...


@guppy(module)
def main(qs: array[qubit, 42] @owned) -> array[qubit, 42]:
    return foo(qs, qs[0])


module.compile()



from hugr.package import ModulePointer

import guppylang.decorator
from guppylang.module import GuppyModule
from guppylang.std.angles import angle

from guppylang.std.builtins import nat, owned
from guppylang.std.qsystem.random import RNG, maybe_rng
from guppylang.std.qsystem.utils import get_current_shot
from guppylang.std.quantum import qubit
from guppylang.std.qsystem.functional import (
    phased_x,
    zz_phase,
    qsystem_functional,
    measure_and_reset,
    zz_max,
    reset,
    rz,
    measure,
    qfree,
)


def compile_qsystem_guppy(fn) -> ModulePointer:  # type: ignore[no-untyped-def]
    """A decorator that combines @guppy with HUGR compilation.

    Modified version of `tests.util.compile_guppy` that loads the qsytem module.
    """
    assert not isinstance(
        fn,
        GuppyModule,
    ), "`@compile_qsystem_guppy` does not support extra arguments."

    module = GuppyModule("module")
    module.load(angle, qubit, get_current_shot, RNG, maybe_rng)  # type: ignore[arg-type]
    module.load_all(qsystem_functional)
    guppylang.decorator.guppy(module)(fn)
    return module.compile()


def test_qsystem(validate):  # type: ignore[no-untyped-def]
    """Compile various operations from the qsystem extension."""

    @compile_qsystem_guppy
    def test(q1: qubit @ owned, q2: qubit @ owned, a1: angle) -> bool:
        shot = get_current_shot()
        q1 = phased_x(q1, a1, a1)
        q1, q2 = zz_phase(q1, q2, a1)
        q1 = rz(q1, a1)
        q1, q2 = zz_max(q1, q2)
        q1, b = measure_and_reset(q1)
        q1 = reset(q1)
        b = measure(q1)
        qfree(q2)
        return b

    validate(test)


def test_qsystem_random(validate):  # type: ignore[no-untyped-def]
    """Compile various operations from the qsystem random extension."""

    @compile_qsystem_guppy
    def test() -> tuple[int, nat, float, int]:
        rng = RNG(42)
        rng2 = maybe_rng(84)
        rng2.unwrap_nothing()
        rint = rng.random_int()
        rnat = rng.random_nat()
        rfloat = rng.random_float()
        rint_bnd = rng.random_int_bounded(100)
        rng.discard()
        return rint, rnat, rfloat, rint_bnd

    validate(test)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")

X = guppy.type_var("X", module=module)


class Generic:
    """Fake Generic type that doesn't check for type var uniqueness."""
    def __class_getitem__(cls, item):
        return cls


@guppy.struct(module)
class MyStruct(Generic[X, X]):
    x: int


module.compile()



from typing import TYPE_CHECKING

from hugr import Wire
from hugr import tys as ht
from hugr.build.function import Function

from guppylang.compiler.cfg_compiler import compile_cfg
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.compiler.hugr_extension import PartialOp
from guppylang.nodes import CheckedNestedFunctionDef

if TYPE_CHECKING:
    from guppylang.definition.function import CheckedFunctionDef


def compile_global_func_def(
    func: "CheckedFunctionDef",
    builder: Function,
    ctx: CompilerContext,
) -> None:
    """Compiles a top-level function definition to Hugr."""
    cfg = compile_cfg(func.cfg, builder, builder.inputs(), ctx)
    builder.set_outputs(*cfg)


def compile_local_func_def(
    func: CheckedNestedFunctionDef,
    dfg: DFContainer,
    ctx: CompilerContext,
) -> Wire:
    """Compiles a local (nested) function definition to Hugr and loads it into a value.

    Returns the wire output of the `LoadFunc` operation.
    """
    assert func.ty.input_names is not None

    # Pick an order for the captured variables
    captured = list(func.captured.values())
    captured_types = [v.ty.to_hugr() for v, _ in captured]

    # Whether the function calls itself recursively.
    recursive = func.name in func.cfg.live_before[func.cfg.entry_bb]

    # Prepend captured variables to the function arguments
    func_ty = func.ty.to_hugr()
    closure_ty = ht.FunctionType([*captured_types, *func_ty.input], func_ty.output)
    func_builder = dfg.builder.define_function(
        func.name, closure_ty.input, closure_ty.output
    )

    # If we have captured variables and the body contains a recursive occurrence of
    # the function itself, then we provide the partially applied function as a local
    # variable
    call_args: list[Wire] = list(func_builder.inputs())
    if len(captured) > 0 and recursive:
        loaded = func_builder.load_function(func_builder, closure_ty)
        partial = func_builder.add_op(
            PartialOp.from_closure(closure_ty, captured_types),
            loaded,
            *func_builder.input_node[: len(captured)],
        )

        call_args.append(partial)
        func.cfg.input_tys.append(func.ty)

        # Compile the CFG
        cfg = compile_cfg(func.cfg, func_builder, call_args, ctx)
        func_builder.set_outputs(*cfg)
    else:
        # Otherwise, we treat the function like a normal global variable
        from guppylang.definition.function import CompiledFunctionDef

        ctx.compiled[func.def_id] = CompiledFunctionDef(
            func.def_id,
            func.name,
            func,
            func.ty,
            {},
            None,
            func.cfg,
            func_builder,
        )
        ctx.worklist.add(func.def_id)  # will compile the CFG later

    # Finally, load the function into the local data-flow graph
    loaded = dfg.builder.load_function(func_builder, closure_ty)
    if len(captured) > 0:
        loaded = dfg.builder.add_op(
            PartialOp.from_closure(closure_ty, captured_types),
            loaded,
            *(dfg[v] for v, _ in captured),
        )

    return loaded



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array


module = GuppyModule("test")


@guppy(module)
def main() -> None:
   xs = array(array(1, 2, 3), array(1, 2, 3))
   ys = xs
   xs[0][0] = 0


module.compile()



import pytest

from tests.util import compile_guppy


def test_if_no_else(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            y += 1
        return y

    validate(foo)


def test_if_else(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            y += 1
        else:
            y -= 1
        return y

    validate(foo)


def test_if_elif(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            y += 1
        elif y > 4:
            y *= 7
        return y

    validate(foo)


def test_if_elif_else(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            y += 1
        elif y > 4:
            y *= 7
        else:
            y = 1337
        return y

    validate(foo)


def test_if_expr(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        return y + 1 if x else 42

    validate(foo)


def test_if_expr_nested(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        a = y + 1 if x else y * y if 0 < y <= 10 else 42
        b = a if a < 5 or (not x and -7 >= a > 42) else -a
        return a if a > b else b

    validate(foo)


def test_if_expr_tuple(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> (int, int, int):
        t = (y + 1 if x else 42), 8, (y * 2 if x else 64)
        return t

    validate(foo)


def test_if_expr_assign_both(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        (z := 5) if x else (z := 42)
        return z

    validate(foo)


def test_if_expr_assign_cond(validate):
    @compile_guppy
    def foo(x: bool) -> bool:
        return z if (z := x) else False

    validate(foo)


def test_if_expr_reassign(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        y = 5
        (y := 1) if x else 6
        6 if x else (y := 2)
        (y := 3) if x else (y := 4)
        return y

    validate(foo)


def test_if_expr_reassign_cond(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        y = 5
        return y if (y := 42) > 5 else 64 if x else y

    validate(foo)


def test_if_expr_double_type_change(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        y = 4
        (y := 1) if (y := x) else (y := 6)
        return y

    validate(foo)


def test_if_return(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            return y
        y *= 32
        return y

    validate(foo)


def test_else_return(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            y += 3
        else:
            y //= 4
            return y
        return y

    validate(foo)


def test_both_return(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        if x:
            y += 3
            return y
        else:
            y //= 4
            return y

    validate(foo)


def test_nested_return(validate):
    @compile_guppy
    def foo(x: int, y: int) -> int:
        if x > 5:
            if y == 4:
                x *= 4
            else:
                return y
        return x

    validate(foo)


def test_return_defined1(validate):
    @compile_guppy
    def foo(x: int, y: int) -> int:
        if x > 5:
            return y
        else:
            z = 5
        return z

    validate(foo)


def test_return_defined2(validate):
    @compile_guppy
    def foo(x: int) -> int:
        if x > 5:
            z = 45
        else:
            return x
        return z

    validate(foo)


def test_break_different_types1(validate):
    @compile_guppy
    def foo(x: int) -> int:
        z = 0
        while True:
            if x > 5:
                z = False
                break
            else:
                z = 8
            z += x
        return 0

    validate(foo)


def test_break_different_types2(validate):
    @compile_guppy
    def foo(x: int) -> int:
        z = 0
        while True:
            if x > 5:
                z = 8
            else:
                z = True
                break
            z += x
        return 0

    validate(foo)


@pytest.mark.skip("Known bug, `z` redefined with different type")
def test_continue_different_types1(validate):
    @compile_guppy
    def foo(x: int) -> int:
        z = 0
        while True:
            if x > 5:
                z = False
                continue
            else:
                z = 8
            z += x
        return z

    validate(foo)


@pytest.mark.skip("Known bug, `z` redefined with different type")
def test_continue_different_types2(validate):
    @compile_guppy
    def foo(x: int) -> int:
        z = 0
        while True:
            if x > 5:
                z = 8
            else:
                z = False
                continue
            z += x
        return z

    validate(foo)


def test_eq(validate):
    @compile_guppy
    def foo() -> bool:
        x = True
        y = not x
        return x == y

    validate(foo)


def test_zero_as_bool(validate):

    @compile_guppy
    def foo() -> bool:
        if 0:
            return False
        else:
            return True

    validate(foo)


def test_one_as_bool(validate):

    @compile_guppy
    def foo() -> bool:
        if 1:
            return True
        else:
            return False

    validate(foo)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array

module = GuppyModule("test")

S = guppy.type_var("S", module=module)
T = guppy.type_var("T", module=module)
U = guppy.type_var("U", module=module)


@guppy.declare(module)
def foo(x: int) -> None:
    ...


@guppy(module)
def main() -> None:
    foo[int](0)


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

module = GuppyModule("test")


@guppy(module)
def f(x: int) -> int:
    return x


@guppy(module)
def g(x: int) -> int:
    return x


@guppy(module)
def main() -> tuple[int, int]:
    return (f, g)(1, 2)


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array


module = GuppyModule("test")

@guppy(module)
def foo(xs: array[array[int, 10], 20]) -> array[int, 10]:
    return xs[0]

module.compile()



from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned


def test_id(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy(module)
    def identity(x: T) -> T:
        return x

    validate(module.compile())


def test_nonlinear(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy(module)
    def copy(x: T) -> tuple[T, T]:
        return x, x

    validate(module.compile())


def test_apply(validate):
    module = GuppyModule("test")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy(module)
    def apply(f: Callable[[S], T], x: S) -> T:
        return f(x)

    validate(module.compile())


def test_annotate(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy(module)
    def identity(x: T) -> T:
        y: T = x
        return y

    validate(module.compile())


def test_recurse(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy(module)
    def empty() -> T:
        return empty()

    validate(module.compile())


def test_call(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy(module)
    def identity(x: T) -> T:
        return x

    @guppy(module)
    def main() -> float:
        return identity(5) + identity(42.0)

    validate(module.compile())


def test_nat(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)
    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def foo(xs: array[T, n] @ owned) -> array[T, n]:
        return xs

    validate(module.compile())


def test_nat_use(validate):
    module = GuppyModule("test")
    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def foo(xs: array[int, n]) -> int:
        return int(n)

    validate(module.compile())


def test_nat_call(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)
    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def foo() -> array[T, n]:
        return foo()

    @guppy(module)
    def main() -> tuple[array[int, 10], array[float, 20]]:
        return foo(), foo()

    validate(module.compile())


def test_nat_recurse(validate):
    module = GuppyModule("test")
    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def empty() -> array[int, n]:
        return empty()

    validate(module.compile())


def test_type_apply(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)
    n = guppy.nat_var("n", module=module)

    @guppy.declare(module)
    def foo(x: array[T, n]) -> array[T, n]: ...

    @guppy(module)
    def identity(x: array[T, n]) -> array[T, n]:
        return foo[T, n](x)

    validate(module.compile())




from pytket import Circuit

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

circ = Circuit(2)
circ.X(0)
circ.Y(1)

module = GuppyModule("test")
module.load(qubit)

@guppy.pytket(circ, module)
def guppy_circ(q: qubit) -> None: ...

@guppy(module)
def foo(q: qubit) -> None:
    guppy_circ(q)


module.compile()


import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct:
    q: qubit


@guppy(module)
def foo(xs: list[int], s: MyStruct @owned) -> list[MyStruct]:
    return [s for x in xs]


module.compile()



from __future__ import annotations

from typing import TYPE_CHECKING

import guppylang
from guppylang.module import GuppyModule

if TYPE_CHECKING:
    from hugr.package import FuncDefnPointer, PackagePointer


def compile_guppy(fn) -> FuncDefnPointer:
    """A decorator that combines @guppy with HUGR compilation.

    Creates a temporary module that only contains the defined function.
    """
    assert not isinstance(
        fn,
        GuppyModule,
    ), "`@compile_guppy` does not support extra arguments."

    module = GuppyModule("module")
    defn = guppylang.decorator.guppy(module)(fn)
    return defn.compile()


def dump_llvm(package: PackagePointer):
    try:
        from execute_llvm import compile_module_to_string

        llvm_module = compile_module_to_string(package.package.to_bytes())
        print(llvm_module)  # noqa: T201

    except ImportError:
        pass



import ast
import copy
import itertools
from collections.abc import Iterator
from dataclasses import dataclass
from typing import ClassVar, NamedTuple

from guppylang.ast_util import (
    AstVisitor,
    ContextAdjuster,
    find_nodes,
    set_location_from,
    template_replace,
    with_loc,
)
from guppylang.cfg.bb import BB, BBStatement
from guppylang.cfg.cfg import CFG
from guppylang.checker.core import Globals
from guppylang.checker.errors.generic import ExpectedError, UnsupportedError
from guppylang.diagnostic import Error
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.experimental import check_lists_enabled
from guppylang.nodes import (
    DesugaredGenerator,
    DesugaredGeneratorExpr,
    DesugaredListComp,
    IterNext,
    MakeIter,
    NestedFunctionDef,
    PyExpr,
)
from guppylang.tys.ty import NoneType

# In order to build expressions, need an endless stream of unique temporary variables
# to store intermediate results
tmp_vars: Iterator[str] = (f"%tmp{i}" for i in itertools.count())


def is_tmp_var(x: str) -> bool:
    """Checks if a name corresponds to a temporary variable."""
    return x.startswith("%tmp")


class Jumps(NamedTuple):
    """Holds jump targets for return, continue, and break during CFG construction."""

    return_bb: BB
    continue_bb: BB | None
    break_bb: BB | None


@dataclass(frozen=True)
class UnreachableError(Error):
    title: ClassVar[str] = "Unreachable"
    span_label: ClassVar[str] = "This code is not reachable"


class CFGBuilder(AstVisitor[BB | None]):
    """Constructs a CFG from ast nodes."""

    cfg: CFG
    globals: Globals

    def build(self, nodes: list[ast.stmt], returns_none: bool, globals: Globals) -> CFG:
        """Builds a CFG from a list of ast nodes.

        We also require the expected number of return ports for the whole CFG. This is
        needed to translate return statements into assignments of dummy return
        variables.
        """
        self.cfg = CFG()
        self.globals = globals

        final_bb = self.visit_stmts(
            nodes, self.cfg.entry_bb, Jumps(self.cfg.exit_bb, None, None)
        )

        # Compute reachable BBs
        self.cfg.update_reachable()

        # If we're still in a basic block after compiling the whole body, we have to add
        # an implicit void return
        if final_bb is not None:
            self.cfg.link(final_bb, self.cfg.exit_bb)
            if final_bb.reachable:
                self.cfg.exit_bb.reachable = True
                if not returns_none:
                    raise GuppyError(ExpectedError(nodes[-1], "return statement"))

        # Prune the CFG such that there are no jumps from unreachable code back into
        # reachable code. Otherwise, unreachable code could lead to unnecessary type
        # checking errors, e.g. if unreachable code changes the type of a variable.
        for bb in self.cfg.bbs:
            if not bb.reachable:
                for succ in list(bb.successors):
                    if succ.reachable:
                        bb.successors.remove(succ)
                        succ.predecessors.remove(bb)
            # Similarly, if a BB is reachable, then there is no need to hold on to dummy
            # jumps into it. Dummy jumps are only needed to propagate type information
            # into and between unreachable BBs
            else:
                for pred in bb.dummy_predecessors:
                    pred.dummy_successors.remove(bb)
                bb.dummy_predecessors = []

        return self.cfg

    def visit_stmts(self, nodes: list[ast.stmt], bb: BB, jumps: Jumps) -> BB | None:
        prev_bb = bb
        bb_opt: BB | None = bb
        next_functional = False
        for node in nodes:
            # If the previous statement jumped, then all following statements are
            # unreachable. Just create a new dummy BB and keep going so we can still
            # check the unreachable code.
            if bb_opt is None:
                bb_opt = self.cfg.new_bb()
                self.cfg.dummy_link(prev_bb, bb_opt)
            if is_functional_annotation(node):
                next_functional = True
                continue

            if next_functional:
                # TODO: This should be an assertion that the Hugr can be un-flattened
                raise NotImplementedError
                next_functional = False
            else:
                prev_bb, bb_opt = bb_opt, self.visit(node, bb_opt, jumps)
        return bb_opt

    def _build_node_value(self, node: BBStatement, bb: BB) -> BB:
        """Utility method for building a node containing a `value` expression.

        Builds the expression and mutates `node.value` to point to the built expression.
        Returns the BB in which the expression is available and adds the node to it.
        """
        if not isinstance(node, NestedFunctionDef) and node.value is not None:
            node.value, bb = ExprBuilder.build(node.value, self.cfg, bb)
        bb.statements.append(node)
        return bb

    def visit_Assign(self, node: ast.Assign, bb: BB, jumps: Jumps) -> BB | None:
        return self._build_node_value(node, bb)

    def visit_AugAssign(self, node: ast.AugAssign, bb: BB, jumps: Jumps) -> BB | None:
        return self._build_node_value(node, bb)

    def visit_AnnAssign(self, node: ast.AnnAssign, bb: BB, jumps: Jumps) -> BB | None:
        return self._build_node_value(node, bb)

    def visit_Expr(self, node: ast.Expr, bb: BB, jumps: Jumps) -> BB | None:
        # This is an expression statement where the value is discarded
        node.value, bb = ExprBuilder.build(node.value, self.cfg, bb)
        # We don't add it to the BB if it's just a temporary variable. This will be the
        # case if it's a branching expression, e.g. `42 if cond else False`. In that
        # example the type mismatch is actually fine since the result is never used. To
        # achieve this behaviour we must not add the temporary result variable to the BB
        if not isinstance(node.value, ast.Name) or not is_tmp_var(node.value.id):
            bb.statements.append(node)
        return bb

    def visit_If(self, node: ast.If, bb: BB, jumps: Jumps) -> BB | None:
        then_bb, else_bb = self.cfg.new_bb(), self.cfg.new_bb()
        BranchBuilder.add_branch(node.test, self.cfg, bb, then_bb, else_bb)
        then_bb = self.visit_stmts(node.body, then_bb, jumps)
        else_bb = self.visit_stmts(node.orelse, else_bb, jumps)
        # We need to handle different cases depending on whether branches jump (i.e.
        # return, continue, or break)
        if then_bb is None:
            # If branch jumps: We continue in the BB of the else branch
            return else_bb
        elif else_bb is None:
            # Else branch jumps: We continue in the BB of the if branch
            return then_bb
        else:
            # No branch jumps: We have to merge the control flow
            return self.cfg.new_bb(then_bb, else_bb)

    def visit_While(self, node: ast.While, bb: BB, jumps: Jumps) -> BB | None:
        head_bb = self.cfg.new_bb(bb)
        body_bb, tail_bb = self.cfg.new_bb(), self.cfg.new_bb()
        BranchBuilder.add_branch(node.test, self.cfg, head_bb, body_bb, tail_bb)

        new_jumps = Jumps(
            return_bb=jumps.return_bb, continue_bb=head_bb, break_bb=tail_bb
        )
        body_end_bb = self.visit_stmts(node.body, body_bb, new_jumps)

        # Go back to the head (but only the body doesn't do its jumping)
        if body_end_bb is not None:
            self.cfg.link(body_end_bb, head_bb)

        # Continue compilation in the tail. This should even happen if the body does
        # its own jumps since the body is not guaranteed to execute
        return tail_bb

    def visit_For(self, node: ast.For, bb: BB, jumps: Jumps) -> BB | None:
        template = """
            it = make_iter
            while True:
                res = iter_next
                if not res.is_some():
                    res.unwrap_nothing()
                    break
                x, it = res.unwrap()
                body
        """

        it = make_var(next(tmp_vars), node.iter)
        res = make_var(next(tmp_vars), node.iter)
        new_nodes = template_replace(
            template,
            node.iter,
            it=it,
            res=res,
            x=node.target,
            make_iter=with_loc(node.iter, MakeIter(value=node.iter, origin_node=node)),
            iter_next=with_loc(node.iter, IterNext(value=it)),
            body=node.body,
        )
        return self.visit_stmts(new_nodes, bb, jumps)

    def visit_Continue(self, node: ast.Continue, bb: BB, jumps: Jumps) -> BB | None:
        if not jumps.continue_bb:
            raise InternalGuppyError("Continue BB not defined")
        self.cfg.link(bb, jumps.continue_bb)
        return None

    def visit_Break(self, node: ast.Break, bb: BB, jumps: Jumps) -> BB | None:
        if not jumps.break_bb:
            raise InternalGuppyError("Break BB not defined")
        self.cfg.link(bb, jumps.break_bb)
        return None

    def visit_Return(self, node: ast.Return, bb: BB, jumps: Jumps) -> BB | None:
        bb = self._build_node_value(node, bb)
        self.cfg.link(bb, jumps.return_bb)
        return None

    def visit_Pass(self, node: ast.Pass, bb: BB, jumps: Jumps) -> BB | None:
        return bb

    def visit_FunctionDef(
        self, node: ast.FunctionDef, bb: BB, jumps: Jumps
    ) -> BB | None:
        from guppylang.checker.func_checker import (
            check_signature,
            parse_function_with_docstring,
        )

        node, docstring = parse_function_with_docstring(node)

        func_ty = check_signature(node, self.globals)
        returns_none = isinstance(func_ty.output, NoneType)
        cfg = CFGBuilder().build(node.body, returns_none, self.globals)

        new_node = NestedFunctionDef(
            cfg,
            func_ty,
            docstring=docstring,
            name=node.name,
            args=node.args,
            body=node.body,
            decorator_list=node.decorator_list,
            returns=node.returns,
            type_comment=node.type_comment,
        )
        set_location_from(new_node, node)
        bb.statements.append(new_node)
        return bb

    def generic_visit(self, node: ast.AST, bb: BB, jumps: Jumps) -> BB | None:
        # When adding support for new statements, we have to remember to use the
        # ExprBuilder to transform all included expressions!
        raise GuppyError(UnsupportedError(node, "This statement", singular=True))


class ExprBuilder(ast.NodeTransformer):
    """Builds an expression into a basic block."""

    cfg: CFG
    bb: BB

    def __init__(self, cfg: CFG, start_bb: BB) -> None:
        self.cfg = cfg
        self.bb = start_bb

    @staticmethod
    def build(node: ast.expr, cfg: CFG, bb: BB) -> tuple[ast.expr, BB]:
        """Builds an expression into a CFG.

        The expression may be transformed and new basic blocks may be created (for
        example for `... if ... else ...` expressions). Returns the new expression and
        the final basic block in which the expression can be used."""
        builder = ExprBuilder(cfg, bb)
        return builder.visit(node), builder.bb

    @classmethod
    def _tmp_assign(cls, tmp_name: str, value: ast.expr, bb: BB) -> None:
        """Adds a temporary variable assignment to a basic block."""
        lhs = make_var(tmp_name, value)
        bb.statements.append(make_assign([lhs], value))

    def visit_Name(self, node: ast.Name) -> ast.Name:
        return node

    def visit_NamedExpr(self, node: ast.NamedExpr) -> ast.Name:
        # This is an assignment expression, e.g. `x := 42`. We turn it into an
        # assignment statement and replace the expression with `x`.
        if not isinstance(node.target, ast.Name):
            raise InternalGuppyError(f"Unexpected assign target: {node.target}")
        assign = ast.Assign(
            targets=[copy.deepcopy(node.target)], value=self.visit(node.value)
        )
        set_location_from(assign, node)
        self.bb.statements.append(assign)
        return node.target

    def visit_IfExp(self, node: ast.IfExp) -> ast.Name:
        if_bb, else_bb = self.cfg.new_bb(), self.cfg.new_bb()
        BranchBuilder.add_branch(node.test, self.cfg, self.bb, if_bb, else_bb)

        if_expr, if_bb = self.build(node.body, self.cfg, if_bb)
        else_expr, else_bb = self.build(node.orelse, self.cfg, else_bb)

        # Assign the result to a temporary variable
        tmp = next(tmp_vars)
        self._tmp_assign(tmp, if_expr, if_bb)
        self._tmp_assign(tmp, else_expr, else_bb)

        # Merge the temporary variables in a new BB
        merge_bb = self.cfg.new_bb(if_bb, else_bb)
        self.bb = merge_bb

        # The final value is stored in the temporary variable
        return make_var(tmp, node)

    def visit_ListComp(self, node: ast.ListComp) -> DesugaredListComp:
        check_lists_enabled(node)
        generators, elt = desugar_comprehension(node.generators, node.elt, node)
        return with_loc(node, DesugaredListComp(elt=elt, generators=generators))

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> DesugaredGeneratorExpr:
        generators, elt = desugar_comprehension(node.generators, node.elt, node)
        return with_loc(node, DesugaredGeneratorExpr(elt=elt, generators=generators))

    def visit_Call(self, node: ast.Call) -> ast.AST:
        return is_py_expression(node) or self.generic_visit(node)

    def generic_visit(self, node: ast.AST) -> ast.AST:
        # Short-circuit expressions must be built using the `BranchBuilder`. However, we
        # can turn them into regular expressions by assigning True/False to a temporary
        # variable and merging the control-flow
        if is_short_circuit_expr(node):
            assert isinstance(node, ast.expr)
            true_bb, false_bb = self.cfg.new_bb(), self.cfg.new_bb()
            BranchBuilder.add_branch(node, self.cfg, self.bb, true_bb, false_bb)
            true_const = ast.Constant(value=True)
            false_const = ast.Constant(value=False)
            set_location_from(true_const, node)
            set_location_from(false_const, node)
            tmp = next(tmp_vars)
            self._tmp_assign(tmp, true_const, true_bb)
            self._tmp_assign(tmp, false_const, false_bb)
            merge_bb = self.cfg.new_bb(true_bb, false_bb)
            self.bb = merge_bb
            return make_var(tmp, node)
        # For all other expressions, just recurse deeper with the node transformer
        return super().generic_visit(node)


class BranchBuilder(AstVisitor[None]):
    """Builds an expression and does branching based on the value.

    This builder should be used to handle all branching on boolean values since it
    handles short-circuit evaluation etc.
    """

    cfg: CFG

    def __init__(self, cfg: CFG):
        """Creates a new `BranchBuilder`."""
        self.cfg = cfg

    @staticmethod
    def add_branch(node: ast.expr, cfg: CFG, bb: BB, true_bb: BB, false_bb: BB) -> None:
        """Builds an expression and branches to `true_bb` or `false_bb`, depending on
        the truth value of the expression."""
        builder = BranchBuilder(cfg)
        builder.visit(node, bb, true_bb, false_bb)

    def visit_Constant(
        self, node: ast.Constant, bb: BB, true_bb: BB, false_bb: BB
    ) -> None:
        # Branching on `True` or `False` constant should be unconditional
        if isinstance(node.value, bool):
            self.cfg.link(bb, true_bb if node.value else false_bb)
            self.cfg.dummy_link(bb, false_bb if node.value else true_bb)
        else:
            self.generic_visit(node, bb, true_bb, false_bb)

    def visit_BoolOp(self, node: ast.BoolOp, bb: BB, true_bb: BB, false_bb: BB) -> None:
        # Add short-circuit evaluation of boolean expression. If there are more than 2
        # operators, we turn the flat operator list into a right-nested tree to allow
        # for recursive processing.
        assert len(node.values) > 1
        if len(node.values) > 2:
            r = ast.BoolOp(
                op=node.op,
                values=node.values[1:],
                lineno=node.values[0].lineno,
                col_offset=node.values[0].col_offset,
                end_lineno=node.values[-1].end_lineno,
                end_col_offset=node.values[-1].end_col_offset,
            )
            node.values = [node.values[0], r]
        [left, right] = node.values

        extra_bb = self.cfg.new_bb()
        assert type(node.op) in [ast.And, ast.Or]
        if isinstance(node.op, ast.And):
            self.visit(left, bb, extra_bb, false_bb)
        elif isinstance(node.op, ast.Or):
            self.visit(left, bb, true_bb, extra_bb)
        self.visit(right, extra_bb, true_bb, false_bb)

    def visit_UnaryOp(
        self, node: ast.UnaryOp, bb: BB, true_bb: BB, false_bb: BB
    ) -> None:
        # For `not` operator, we can just switch `true_bb` and `false_bb`
        if isinstance(node.op, ast.Not):
            self.visit(node.operand, bb, false_bb, true_bb)
        else:
            self.generic_visit(node, bb, true_bb, false_bb)

    def visit_Compare(
        self, node: ast.Compare, bb: BB, true_bb: BB, false_bb: BB
    ) -> None:
        # Support chained comparisons, e.g. `x <= 5 < y` by compiling to `x <= 5 and
        # 5 < y`. This way we get short-circuit evaluation for free.
        if len(node.comparators) > 1:
            comparators = [node.left, *node.comparators]
            values = [
                ast.Compare(
                    left=left,
                    ops=[op],
                    comparators=[right],
                    lineno=left.lineno,
                    col_offset=left.col_offset,
                    end_lineno=right.end_lineno,
                    end_col_offset=right.end_col_offset,
                )
                for left, op, right in zip(
                    comparators[:-1], node.ops, comparators[1:], strict=True
                )
            ]
            conj = ast.BoolOp(op=ast.And(), values=values)
            set_location_from(conj, node)
            self.visit_BoolOp(conj, bb, true_bb, false_bb)
        else:
            self.generic_visit(node, bb, true_bb, false_bb)

    def visit_IfExp(self, node: ast.IfExp, bb: BB, true_bb: BB, false_bb: BB) -> None:
        then_bb, else_bb = self.cfg.new_bb(), self.cfg.new_bb()
        self.visit(node.test, bb, then_bb, else_bb)
        self.visit(node.body, then_bb, true_bb, false_bb)
        self.visit(node.orelse, else_bb, true_bb, false_bb)

    def generic_visit(self, node: ast.expr, bb: BB, true_bb: BB, false_bb: BB) -> None:
        # We can always fall back to building the node as a regular expression and using
        # the result as a branch predicate
        pred, bb = ExprBuilder.build(node, self.cfg, bb)
        bb.branch_pred = pred
        self.cfg.link(bb, false_bb)
        self.cfg.link(bb, true_bb)


def desugar_comprehension(
    generators: list[ast.comprehension], elt: ast.expr, node: ast.AST
) -> tuple[list[DesugaredGenerator], ast.expr]:
    """Helper function to desugar a comprehension node."""
    # Check for illegal expressions
    illegals = find_nodes(is_illegal_in_list_comp, node)
    if illegals:
        err = UnsupportedError(
            illegals[0],
            "This expression",
            singular=True,
            unsupported_in="a list comprehension",
        )
        raise GuppyError(err)

    # The check above ensures that the comprehension doesn't contain any control-flow
    # expressions. Thus, we can use a dummy `ExprBuilder` to desugar the insides.
    # TODO: Refactor so that desugaring is separate from control-flow building
    dummy_cfg = CFG()
    builder = ExprBuilder(dummy_cfg, dummy_cfg.entry_bb)

    # Desugar into statements that create the iterator, check for a next element,
    # get the next element, and finalise the iterator.
    gens = []
    for g in generators:
        if g.is_async:
            raise GuppyError(UnsupportedError(g, "Async generators"))
        g.iter = builder.visit(g.iter)
        it = make_var(next(tmp_vars), g.iter)
        desugared = DesugaredGenerator(
            iter=it,
            iter_assign=make_assign(
                [it], with_loc(it, MakeIter(value=g.iter, origin_node=node))
            ),
            next_call=with_loc(it, IterNext(value=it)),
            target=g.target,
            ifs=g.ifs,
            borrowed_outer_places=[],
        )
        gens.append(desugared)

    elt = builder.visit(elt)
    return gens, elt


def is_functional_annotation(stmt: ast.stmt) -> bool:
    """Returns `True` iff the given statement is the functional pseudo-decorator.

    Pseudo-decorators are built using the matmul operator `@`, i.e. `_@functional`.
    """
    if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.BinOp):
        op = stmt.value
        if (
            isinstance(op.op, ast.MatMult)
            and isinstance(op.left, ast.Name)
            and isinstance(op.right, ast.Name)
        ):
            return op.left.id == "_" and op.right.id == "functional"
    return False


@dataclass(frozen=True)
class EmptyPyExprError(Error):
    title: ClassVar[str] = "Invalid Python expression"
    span_label: ClassVar[str] = "Compile-time `py(...)` expression requires an argument"


def is_py_expression(node: ast.AST) -> PyExpr | None:
    """Checks if the given node is a compile-time `py(...)` expression and turns it into
    a `PyExpr` AST node.

    Otherwise, returns `None`.
    """
    if (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Name)
        and node.func.id == "py"
    ):
        match node.args:
            case []:
                raise GuppyError(EmptyPyExprError(node))
            case [arg]:
                pass
            case args:
                arg = with_loc(node, ast.Tuple(elts=args, ctx=ast.Load))
        return with_loc(node, PyExpr(value=arg))
    return None


def is_short_circuit_expr(node: ast.AST) -> bool:
    """Checks if an expression uses short-circuiting.

    Those expressions *must* be compiled using the `BranchBuilder`.
    """
    return isinstance(node, ast.BoolOp) or (
        isinstance(node, ast.Compare) and len(node.comparators) > 1
    )


def is_illegal_in_list_comp(node: ast.AST) -> bool:
    """Checks if an expression is illegal to use in a list comprehension."""
    return isinstance(node, ast.IfExp | ast.NamedExpr) or is_short_circuit_expr(node)


def make_var(name: str, loc: ast.AST | None = None) -> ast.Name:
    """Creates an `ast.Name` node."""
    node = ast.Name(id=name, ctx=ast.Load)
    if loc is not None:
        set_location_from(node, loc)
    return node


def make_assign(lhs: list[ast.AST], value: ast.expr) -> ast.Assign:
    """Creates an `ast.Assign` node."""
    assert len(lhs) > 0
    adjuster = ContextAdjuster(ast.Store())
    lhs = [adjuster.visit(expr) for expr in lhs]
    if len(lhs) == 1:
        target = lhs[0]
    else:
        target = with_loc(value, ast.Tuple(elts=lhs, ctx=ast.Store()))
    return with_loc(value, ast.Assign(targets=[target], value=value))



from guppylang.tys.builtin import array_type_def
from guppylang.tys.param import ConstParam, TypeParam
from guppylang.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NumericType,
    OpaqueType,
)


def test_generic_function_type():
    ty_param = TypeParam(0, "T", must_be_copyable=False, must_be_droppable=False)
    len_param = ConstParam(1, "n", NumericType(NumericType.Kind.Nat))
    array_ty = OpaqueType([ty_param.to_bound(0), len_param.to_bound(1)], array_type_def)
    ty = FunctionType(
        params=[ty_param, len_param],
        inputs=[FuncInput(array_ty, InputFlags.Inout)],
        output=ty_param.to_bound(0).ty,
    )
    assert str(ty) == "forall T, n: nat. array[T, n] -> T"



import pytest

from hugr import ops
from hugr.std.int import IntVal

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned, mem_swap, nat
from tests.util import compile_guppy

from guppylang.std.quantum import qubit, discard
import guppylang.std.quantum as quantum


@pytest.mark.skip("Requires `is_to_u` in llvm")
def test_len_execute(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main(xs: array[float, 42]) -> int:
        return len(xs)

    compiled = module.compile()
    validate(compiled)
    if run_int_fn is not None:
        run_int_fn(compiled, 42)


def test_len(validate):
    test_len_execute(validate, None)


def test_len_linear(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def main(qs: array[qubit, 42]) -> int:
        return len(qs)

    validate(module.compile())


def test_len_generic(validate):
    module = GuppyModule("test")

    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def main(qs: array[bool, n]) -> bool:
        for i in range(len(qs)):
            if qs[i]:
                return True
        return False

    package = module.compile()
    validate(package)

    hg = package.module
    load_nats = [
        data.op
        for _, data in hg.nodes()
        if isinstance(data.op, ops.ExtOp) and data.op.op_def().name == "load_nat"
    ]
    assert len(load_nats) == 1


def test_index(validate):
    @compile_guppy
    def main(xs: array[int, 5], i: int) -> int:
        return xs[0] + xs[i] + xs[xs[2 * i]]

    validate(main)


def test_new_array(validate):
    @compile_guppy
    def main(x: int, y: int) -> array[int, 3]:
        xs = array(x, y, x)
        return xs

    validate(main)


def test_new_array_infer_empty(validate):
    @compile_guppy
    def main() -> array[float, 0]:
        return array()

    validate(main)


def test_new_array_infer_nested(validate):
    @compile_guppy
    def main(ys: array[int, 0] @ owned) -> array[array[int, 0], 2]:
        xs = array(ys, array())
        return xs

    validate(main)


def test_return_linear_array(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def foo() -> array[qubit, 2]:
        a = array(qubit(), qubit())
        return a

    validate(module.compile())


def test_subscript_drop_rest(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo() -> array[int, 10]: ...

    @guppy(module)
    def main() -> int:
        return foo()[0]

    validate(module.compile())


def test_linear_subscript(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def main(qs: array[qubit, 42] @ owned, i: int) -> array[qubit, 42]:
        foo(qs[i])
        return qs

    validate(module.compile())


def test_inout_subscript(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def main(qs: array[qubit, 42], i: int) -> None:
        foo(qs[i])

    validate(module.compile())


def test_multi_subscripts(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q1: qubit, q2: qubit) -> None: ...

    @guppy(module)
    def main(qs: array[qubit, 42] @ owned) -> array[qubit, 42]:
        foo(qs[0], qs[1])
        foo(qs[0], qs[0])  # Will panic at runtime
        return qs

    validate(module.compile())


def test_struct_array(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class S:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def foo(q1: qubit, q2: qubit) -> None: ...

    @guppy(module)
    def main(ss: array[S, 10] @ owned) -> array[S, 10]:
        # This will panic at runtime :(
        # To make this work, we would need to replace the qubits in the struct
        # with `qubit | None` and write back `None` after `q1` has been extracted...
        foo(ss[0].q1, ss[0].q2)
        return ss

    validate(module.compile())


def test_nested_subscripts(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy.declare(module)
    def bar(q1: qubit, q2: qubit, q3: qubit, q4: qubit) -> None: ...

    @guppy(module)
    def main(qs: array[array[qubit, 13], 42] @ owned) -> array[array[qubit, 13], 42]:
        foo(qs[0][0])
        # The following should work *without* panicking at runtime! Accessing `qs[0][0]`
        # replaces one qubit with `None` but puts everything back into `qs` before
        # going to the next argument.
        bar(qs[0][0], qs[0][1], qs[1][0], qs[1][1])
        return qs

    validate(module.compile())


def test_struct_nested_subscript(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class C:
        c: qubit
        blah: int

    @guppy.struct(module)
    class B:
        ys: array[array[C, 10], 20]
        foo: C

    @guppy.struct(module)
    class A:
        xs: array[B, 42]
        bar: qubit
        baz: tuple[B, B]

    @guppy.declare(module)
    def foo(q1: qubit) -> None: ...

    @guppy(module)
    def main(a: A @ owned, i: int, j: int, k: int) -> A:
        foo(a.xs[i].ys[j][k].c)
        return a

    validate(module.compile())


def test_generic_function(validate):
    module = GuppyModule("test")
    module.load(qubit)
    T = guppy.type_var("T", copyable=False, droppable=False, module=module)
    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def foo(xs: array[T, n] @ owned) -> array[T, n]:
        return xs

    @guppy(module)
    def main() -> tuple[array[int, 3], array[qubit, 2]]:
        xs = array(1, 2, 3)
        ys = array(qubit(), qubit())
        return foo(xs), foo(ys)

    validate(module.compile())


def test_linear_for_loop(validate):
    module = GuppyModule("test")
    module.load(qubit, discard)

    @guppy(module)
    def main() -> None:
        qs = array(qubit(), qubit(), qubit())
        for q in qs:
            discard(q)

    validate(module.compile())


def test_exec_array(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        a = array(1, 2, 3)
        return a[0] + a[1] + a[2]

    package = module.compile()
    validate(package)
    run_int_fn(package, expected=6)


def test_exec_array_loop(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(1, 2, 3, 4, 5, 6, 7)
        s = 0
        for x in xs:
            if x % 2 == 0:
                continue
            if x > 5:
                break
            s += x
        return s

    package = module.compile()
    validate(package)

    run_int_fn(package, expected=9)


def test_mem_swap(validate):
    module = GuppyModule("test")

    module.load(qubit)

    @guppy(module)
    def foo(x: qubit, y: qubit) -> None:
        mem_swap(x, y)

    @guppy(module)
    def main() -> array[qubit, 2]:
        a = array(qubit(), qubit())
        foo(a[0], a[1])
        return a

    package = module.compile()
    validate(package)


def test_drop(validate):
    @compile_guppy
    def main(xs: array[int, 2] @ owned) -> None:
        ys = xs

    validate(main)


def test_copy1(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(1, 2, 3)
        ys = xs.copy()
        xs = array(4, 5, 6)
        return xs[0] + ys[0]  # Check copy isn't modified

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=5)


def test_copy2(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(1, 2, 3)
        ys = copy(xs)
        xs = array(4, 5, 6)
        return xs[0] + ys[0]  # Check copy isn't modified

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=5)


def test_copy3(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(1, 2, 3)
        ys = copy(xs)
        return xs[0]  # Check original can keep being used

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=1)


def test_copy_struct(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class S:
        a: array[int, 1]

    @guppy(module)
    def main() -> int:
        xs = array(S(array(1)), S(array(2)))
        ys = copy(xs[0].a)
        return ys[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=1)


def test_copy_const(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = copy(array(1, 2, 3))
        return xs[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=1)


def test_subscript_assign(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def foo(xs: array[int, 3] @ owned, idx: int, n: int) -> array[int, 3]:
        xs[idx] = n
        return xs

    @guppy(module)
    def main() -> int:
        xs = array(0, 0, 0)
        xs = foo(xs, 0, 2)
        return xs[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=2)


def test_subscript_assign_add(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def foo(xs: array[int, 1], n: int) -> None:
        xs[0] += n

    @guppy(module)
    def main() -> int:
        xs = array(0)
        for i in range(6):
            foo(xs, i)
        return xs[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=15)


def test_subscript_assign_struct(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class S:
        a: array[int, 2]

    @guppy(module)
    def foo(xs: array[int, 2], idx: int, n: int) -> None:
        xs[idx] = n

    @guppy(module)
    def main() -> int:
        s = S(array(0, 0))
        foo(s.a, 1, 42)
        return s.a[1]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=42)


def test_subscript_assign_nested(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def foo(xs: array[array[int, 2], 2]) -> None:
        xs[0][0] = 22

    @guppy(module)
    def main() -> int:
        xs = array(array(11, 11), array(11, 11))
        foo(xs)
        return xs[0][0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=22)


def test_subscript_assign_nested_struct1(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class S:
        a: array[int, 2]

    @guppy(module)
    def main() -> int:
        s0 = S(array(0, 0))
        s1 = S(array(1, 1))
        arr = array(s0, s1)
        arr[0].a[1] = 42
        return arr[0].a[1]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=42)


def test_subscript_assign_nested_struct2(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class A:
        b: array[int, 2]

    @guppy.struct(module)
    class S:
        a: array[A, 2]

    @guppy(module)
    def main() -> int:
        s0 = S(array(A(array(0, 0)), A(array(0, 0))))
        s1 = S(array(A(array(0, 0)), A(array(0, 0))))
        arr = array(s0, s1)
        arr[0].a[1].b = array(42, 42)
        arr[0].a[1].b[0] = 43
        return arr[0].a[1].b[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=43)


def test_subscript_assign_nested_struct3(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class A:
        b: array[int, 2]

    @guppy.struct(module)
    class S:
        a: A

    @guppy(module)
    def main() -> int:
        s0 = S(A(array(0, 0)))
        s1 = S(A(array(0, 0)))
        arr = array(s0, s1)
        arr[0].a.b[0] = 2
        return arr[0].a.b[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=2)


def test_subscript_assign_nested_struct4(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class S:
        xs: array[int, 1]

    @guppy(module)
    def main() -> int:
        arr = array(S(array(0)), S(array(1)))
        arr[0].xs = array(3)
        return arr[0].xs[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=3)


def test_subscript_assign_nested_struct5(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy.struct(module)
    class A:
        b: array[int, 2]

    @guppy.struct(module)
    class S:
        a: array[A, 2]

    @guppy(module)
    def main() -> int:
        s0 = S(array(A(array(0, 0)), A(array(0, 0))))
        s1 = S(array(A(array(0, 0)), A(array(0, 0))))
        arr = array(s0, s1)
        arr[0].a[0].b = array(42, 42)
        return arr[0].a[0].b[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=42)


def test_subscript_assign_unpacking(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(11, 22, 33)
        xs[0], y = 44, 55
        return xs[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=44)


def test_subscript_subscript_assign(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(0)
        ys = array(1)
        xs[0] = ys[0]
        return xs[0]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=1)


def test_subscript_assign_unpacking_tuple(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(0, 0, 0)
        a, *b, xs[1] = (1, 2, 3, 4)
        return xs[1]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=4)


def test_subscript_assign_unpacking_complicated(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(0, 0, 0)
        (a1, a2), *b, (c1, c2) = array((0, 1), (2, 3), (4, 5), (5, 6))
        return a1 + c1

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=5)


def test_subscript_assign_unpacking_range(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(0, 0, 0)
        a, *b, xs[1] = range(10)
        return xs[1]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=9)


def test_subscript_assign_unpacking_array(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        xs = array(0, 0, 0)
        a, *b, xs[1] = array(1, 2, 3, 4)
        return xs[1]

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=4)


# Verifies that array.__getitem__ works across multiple functions calls.
def test_multiple_functions(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def first(arr: array[int, 2]) -> int:
        return arr[0]

    @guppy(module)
    def second(arr: array[int, 2]) -> int:
        return arr[1]

    @guppy(module)
    def main() -> int:
        xs = array(1, 2)
        return first(xs) + second(xs)

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=3)


def test_assign_dataflow(validate):
    """Test that dataflow analysis considers subscript assignments as uses and correctly
    wires up the Hugr.

    See https://github.com/CQCL/guppylang/issues/844
    """
    module = GuppyModule("test")

    @guppy(module)
    def test1() -> None:
        xs = array(1, 2)
        for i in range(2):
            xs[i] = 0

    @guppy(module)
    def test2() -> None:
        xs = array(0, 1)
        ys = array(1, 2)
        for i in range(2):
            ys[xs[i]] = 0

    module.compile()



from typing import Generic, TypeVar

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


X = TypeVar("X")  # This is a Python type variable, not a Guppy one!


@guppy.struct(module)
class MyStruct(Generic[X]):
    x: int


module.compile()



import itertools
from abc import ABC
from collections import defaultdict
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Any, cast

import tket2_exts
from hugr import Hugr, Node, Wire, ops
from hugr import tys as ht
from hugr.build import function as hf
from hugr.build.dfg import DP, DefinitionBuilder, DfBase
from hugr.hugr.base import OpVarCov
from hugr.hugr.node_port import ToNode
from hugr.std import PRELUDE

from guppylang.checker.core import FieldAccess, Globals, Place, PlaceId, Variable
from guppylang.definition.common import CheckedDef, CompilableDef, CompiledDef, DefId
from guppylang.definition.ty import TypeDef
from guppylang.definition.value import CompiledCallableDef
from guppylang.error import InternalGuppyError
from guppylang.tys.ty import StructType, Type

CompiledLocals = dict[PlaceId, Wire]


@dataclass(frozen=True)
class GlobalConstId:
    id: int
    base_name: str

    _fresh_ids = itertools.count()

    @staticmethod
    def fresh(base_name: str) -> "GlobalConstId":
        return GlobalConstId(next(GlobalConstId._fresh_ids), base_name)

    @property
    def name(self) -> str:
        return f"{self.base_name}.{self.id}"


class CompilerContext:
    """Compilation context containing all available definitions.

    Maintains a `worklist` of definitions which have been used by other compiled code
    (i.e. `compile_outer` has been called) but have not yet been compiled/lowered
    themselves (i.e. `compile_inner` has not yet been called).
    """

    module: DefinitionBuilder[ops.Module]
    checked: dict[DefId, CheckedDef]
    compiled: dict[DefId, CompiledDef]
    worklist: set[DefId]

    global_funcs: dict[GlobalConstId, hf.Function]

    checked_globals: Globals

    def __init__(
        self,
        checked: dict[DefId, CheckedDef],
        module: DefinitionBuilder[ops.Module],
        checked_globals: Globals,
    ) -> None:
        self.module = module
        self.checked = checked
        self.worklist = set()
        self.compiled = {}
        self.global_funcs = {}
        self.checked_globals = checked_globals

    def build_compiled_def(self, def_id: DefId) -> CompiledDef:
        """Returns the compiled definitions corresponding to the given ID.

        Might mutate the current Hugr if this definition has never been compiled before.
        """
        if def_id not in self.compiled:
            defn = self.checked[def_id]
            self.compiled[def_id] = self._compile(defn)
            self.worklist.add(def_id)
        return self.compiled[def_id]

    def _compile(self, defn: CheckedDef) -> CompiledDef:
        if isinstance(defn, CompilableDef):
            return defn.compile_outer(self.module)
        return defn

    def compile(self, defn: CheckedDef) -> None:
        """Compiles the given definition and all of its dependencies into the current
        Hugr."""
        if defn.id in self.compiled:
            return

        self.compiled[defn.id] = self._compile(defn)
        self.worklist.add(defn.id)
        while self.worklist:
            next_id = self.worklist.pop()
            with track_hugr_side_effects():
                next_def = self.build_compiled_def(next_id)
                next_def.compile_inner(self)

    def get_instance_func(
        self, ty: Type | TypeDef, name: str
    ) -> CompiledCallableDef | None:
        checked_func = self.checked_globals.get_instance_func(ty, name)
        if checked_func is None:
            return None
        compiled_func = self.build_compiled_def(checked_func.id)
        assert isinstance(compiled_func, CompiledCallableDef)
        return compiled_func

    def declare_global_func(
        self,
        const_id: GlobalConstId,
        func_ty: ht.PolyFuncType,
    ) -> tuple[hf.Function, bool]:
        """
        Creates a function builder for a global function if it doesn't already exist,
        else returns the existing one.
        """
        if const_id in self.global_funcs:
            return self.global_funcs[const_id], True
        func = self.module.define_function(
            name=const_id.name,
            input_types=func_ty.body.input,
            output_types=func_ty.body.output,
            type_params=func_ty.params,
        )
        self.global_funcs[const_id] = func
        return func, False


@dataclass
class DFContainer:
    """A dataflow graph under construction.

    This class is passed through the entire compilation pipeline and stores a builder
    for the dataflow child-graph currently being constructed as well as all live local
    variables. Note that the variable map is mutated in-place and always reflects the
    current compilation state.
    """

    builder: DfBase[ops.DfParentOp]
    locals: CompiledLocals = field(default_factory=dict)

    def __init__(
        self, builder: DfBase[DP], locals: CompiledLocals | None = None
    ) -> None:
        generic_builder = cast(DfBase[ops.DfParentOp], builder)
        if locals is None:
            locals = {}
        self.builder = generic_builder
        self.locals = locals

    def __getitem__(self, place: Place) -> Wire:
        """Constructs a wire for a local place in this DFG.

        Note that this mutates the Hugr since we might need to pack or unpack some
        tuples to obtain a port for places that involve struct fields.
        """
        # First check, if we already have a wire for this place
        if place.id in self.locals:
            return self.locals[place.id]
        # Otherwise, our only hope is that it's a struct value that we can rebuild by
        # packing the wires of its constituting fields
        if not isinstance(place.ty, StructType):
            raise InternalGuppyError(f"Couldn't obtain a port for `{place}`")
        children = [FieldAccess(place, field, None) for field in place.ty.fields]
        child_types = [child.ty.to_hugr() for child in children]
        child_wires = [self[child] for child in children]
        wire = self.builder.add_op(ops.MakeTuple(child_types), *child_wires)[0]
        for child in children:
            if child.ty.linear:
                self.locals.pop(child.id)
        self.locals[place.id] = wire
        return wire

    def __setitem__(self, place: Place, port: Wire) -> None:
        # When assigning a struct value, we immediately unpack it recursively and only
        # store the leaf wires.
        is_return = isinstance(place, Variable) and is_return_var(place.name)
        if isinstance(place.ty, StructType) and not is_return:
            unpack = self.builder.add_op(
                ops.UnpackTuple([t.ty.to_hugr() for t in place.ty.fields]), port
            )
            for field, field_port in zip(place.ty.fields, unpack, strict=True):
                self[FieldAccess(place, field, None)] = field_port
            # If we had a previous wire assigned to this place, we need forget about it.
            # Otherwise, we might use this old value when looking up the place later
            self.locals.pop(place.id, None)
        else:
            self.locals[place.id] = port

    def __contains__(self, place: Place) -> bool:
        return place.id in self.locals

    def __copy__(self) -> "DFContainer":
        # Make a copy of the var map so that mutating the copy doesn't
        # mutate our variable mapping
        return DFContainer(self.builder, self.locals.copy())


class CompilerBase(ABC):
    """Base class for the Guppy compiler."""

    ctx: CompilerContext

    def __init__(self, ctx: CompilerContext) -> None:
        self.ctx = ctx


def return_var(n: int) -> str:
    """Name of the dummy variable for the n-th return value of a function.

    During compilation, we treat return statements like assignments of dummy variables.
    For example, the statement `return e0, e1, e2` is treated like `%ret0 = e0 ; %ret1 =
    e1 ; %ret2 = e2`. This way, we can reuse our existing mechanism for passing of live
    variables between basic blocks."""
    return f"%ret{n}"


def is_return_var(x: str) -> bool:
    """Checks whether the given name is a dummy return variable."""
    return x.startswith("%ret")


QUANTUM_EXTENSION = tket2_exts.quantum()
RESULT_EXTENSION = tket2_exts.result()

#: List of extension ops that have side-effects, identified by their qualified name
EXTENSION_OPS_WITH_SIDE_EFFECTS: list[str] = [
    # Results should be order w.r.t. each other but also w.r.t. panics
    *(op_def.qualified_name() for op_def in RESULT_EXTENSION.operations.values()),
    PRELUDE.get_op("panic").qualified_name(),
    # Qubit allocation and deallocation have the side-effect of changing the number of
    # available free qubits
    QUANTUM_EXTENSION.get_op("QAlloc").qualified_name(),
    QUANTUM_EXTENSION.get_op("QFree").qualified_name(),
    QUANTUM_EXTENSION.get_op("MeasureFree").qualified_name(),
]


def may_have_side_effect(op: ops.Op) -> bool:
    """Checks whether an operation could have a side-effect.

    We need to insert implicit state order edges between these kinds of nodes to ensure
    they are executed in the correct order, even if there is no data dependency.
    """
    match op:
        case ops.ExtOp() as ext_op:
            return ext_op.op_def().qualified_name() in EXTENSION_OPS_WITH_SIDE_EFFECTS
        case ops.Custom(op_name=op_name, extension=extension):
            qualified_name = f"{extension}.{op_name}" if extension else op_name
            return qualified_name in EXTENSION_OPS_WITH_SIDE_EFFECTS
        case ops.Call() | ops.CallIndirect():
            # Conservative choice is to assume that all calls could have side effects.
            # In the future we could inspect the call graph to figure out a more
            # precise answer
            return True
        case _:
            return False


@contextmanager
def track_hugr_side_effects() -> Iterator[None]:
    """Initialises the tracking of nodes with side-effects during Hugr building.

    Ensures that state-order edges are implicitly inserted between side-effectful nodes
    to ensure they are executed in the order they are added.
    """
    # Remember original `Hugr.add_node` method that is monkey-patched below.
    hugr_add_node = Hugr.add_node
    # Last node with potential side effects for each dataflow parent
    prev_node_with_side_effect: defaultdict[Node, Node | None] = defaultdict(
        lambda: None
    )

    def hugr_add_node_with_order(
        self: Hugr[OpVarCov],
        op: ops.Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        """Monkey-patched version of `Hugr.add_node` that takes care of implicitly
        inserting state order edges between operations that could have side-effects.
        """
        new_node = hugr_add_node(self, op, parent, num_outs, metadata)
        if may_have_side_effect(op):
            handle_side_effect(new_node, self)
        return new_node

    def handle_side_effect(node: Node, hugr: Hugr[OpVarCov]) -> None:
        """Performs the actual order-edge insertion, assuming that `node` has a side-
        effect."""
        parent = hugr[node].parent
        if parent is not None:
            if prev := prev_node_with_side_effect[parent]:
                hugr.add_order_link(prev, node)
            else:
                # If this is the first side-effectful op in this DFG, make a recursive
                # call with the parent since the parent is also considered side-
                # effectful now. We shouldn't walk up through function definitions
                # or basic blocks though
                if not isinstance(hugr[parent].op, ops.FuncDefn | ops.DataflowBlock):
                    handle_side_effect(parent, hugr)
            prev_node_with_side_effect[parent] = node

    # Monkey-patch the `add_node` method
    Hugr.add_node = hugr_add_node_with_order  # type: ignore[method-assign]
    try:
        yield
    finally:
        Hugr.add_node = hugr_add_node  # type: ignore[method-assign]



from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std._internal.compiler.qsystem import (
    RandomIntBoundedCompiler,
    RandomIntCompiler,
)
from guppylang.std._internal.compiler.quantum import (
    QSYSTEM_RANDOM_EXTENSION,
    RNGCONTEXT_T,
)
from guppylang.std._internal.util import external_op
from guppylang.std.builtins import nat, owned
from guppylang.std.option import Option

qsystem_random = GuppyModule("qsystem.random")


@guppy.hugr_op(
    external_op("NewRNGContext", [], ext=QSYSTEM_RANDOM_EXTENSION),
    module=qsystem_random,
)
@no_type_check
def _new_rng_context(seed: int) -> Option["RNG"]: ...


@guppy(qsystem_random)
def maybe_rng(seed: int) -> Option["RNG"]:  # type: ignore[type-arg] # "Option" expects no type arguments, but 1 given
    """Safely create a new random number generator using a seed.

    Returns `nothing` if RNG is already initialized."""
    return _new_rng_context(seed)  # type: ignore[no-any-return] # Returning Any from function declared to return "Option"


@guppy.type(RNGCONTEXT_T, copyable=False, droppable=False, module=qsystem_random)
class RNG:
    """Random number generator."""

    @guppy(qsystem_random)  # type: ignore[misc] # Unsupported decorated constructor type; Self argument missing for a non-static method (or an invalid type for self)
    def __new__(seed: int) -> "RNG":
        """Create a new random number generator using a seed."""
        return _new_rng_context(seed).unwrap()  # type: ignore[no-any-return] # Returning Any from function declared to return "RNGContext"

    @guppy.hugr_op(
        external_op("DeleteRNGContext", [], ext=QSYSTEM_RANDOM_EXTENSION),
        module=qsystem_random,
    )
    @no_type_check
    def discard(self: "RNG" @ owned) -> None: ...

    @guppy.custom(RandomIntCompiler(), module=qsystem_random)
    @no_type_check
    def random_int(self: "RNG") -> int: ...

    @guppy(qsystem_random)
    def random_nat(self: "RNG") -> nat:
        """Generate a random 32-bit natural number."""
        return nat(self.random_int())  # type: ignore[call-arg] # Too many arguments for "nat"

    @guppy.hugr_op(
        external_op("RandomFloat", [], ext=QSYSTEM_RANDOM_EXTENSION),
        module=qsystem_random,
    )
    @no_type_check
    def random_float(self: "RNG") -> float: ...

    @guppy.custom(RandomIntBoundedCompiler(), module=qsystem_random)
    @no_type_check
    def random_int_bounded(self: "RNG", bound: int) -> int: ...



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array


module = GuppyModule("test")

@guppy.declare(module)
def foo() -> array[int, 10]: ...

@guppy(module)
def main() -> None:
    foo()[0] = 22

module.compile()


import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct:
    q1: qubit
    q2: qubit


@guppy(module)
def foo() -> MyStruct:
    return MyStruct(qubit(), qubit())


@guppy(module)
def bar() -> qubit:
    return foo().q1


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.tys.ty import NoneType


module = GuppyModule("test")


@guppy.type(NoneType().to_hugr(), module=module)
class MyType:
    """A type where the `__iter__` method has the wrong signature."""

    @guppy.declare(module)
    def __iter__(self: "MyType", x: int) -> "MyType":
        ...


@guppy(module)
def test(x: MyType) -> None:
    for _ in x:
        pass


module.compile()
